//
use crate::core::{
    git::UserNotification,
    server::{EmailConfig, SmtpConfig},
};

use crate::alerts::{full_message, short_message};

#[derive(Debug)]
struct EmailMessage {
    from: String,
    to: String,
    subject: String,
    body: String,
}

pub(super) fn send_email(user_notification: &UserNotification, email_config: &EmailConfig) {
    let email_message = EmailMessage {
        from: email_config.from().to_string(),
        to: email_config.to().to_string(),
        subject: short_message(user_notification),
        body: full_message(user_notification),
    };
    match email_config.smtp() {
        Some(smtp) => send_email_smtp(email_message, smtp),
        None => send_email_sendmail(&email_message),
    }
}

fn send_email_sendmail(email_message: &EmailMessage) {
    use sendmail::email;
    match email::send(
        &email_message.from,
        [email_message.to.as_ref()],
        &email_message.subject,
        &email_message.body,
    ) {
        Ok(()) => tracing::info!("Email sent successfully!"),
        Err(e) => tracing::warn!("Could not send email: {:?}", e),
    }
}

fn send_email_smtp(email_message: EmailMessage, smtp: &SmtpConfig) {
    if let Err(err) = do_send_email_smtp(email_message, smtp) {
        tracing::warn!(?err, "sending email");
    }
}

fn do_send_email_smtp(email_message: EmailMessage, smtp: &SmtpConfig) -> Result<(), anyhow::Error> {
    use lettre::{transport::smtp::authentication::Credentials, Message, SmtpTransport, Transport};
    let email = Message::builder()
        .from(email_message.from.parse()?)
        .to(email_message.to.parse()?)
        .subject(email_message.subject)
        .body(email_message.body)?;
    let creds = Credentials::new(smtp.username().to_string(), smtp.password().to_string());
    let mailer = SmtpTransport::relay(smtp.hostname())?
        .credentials(creds)
        .build();
    Ok(mailer
        .send(&email)
        .map(|response| {
            response
                .message()
                .map(ToString::to_string)
                .collect::<Vec<_>>()
        })
        .map(|response| {
            tracing::info!(?response, "email sent via smtp");
        })?)
}
