//
use crate::core::git::UserNotification;
use tracing::info;

use std::{
    collections::HashMap,
    time::{Duration, Instant},
};

#[derive(Debug)]
pub struct History {
    /// The maximum age of an item in the history.
    ///
    /// Items older than this will be dropped.
    max_age_seconds: Duration,

    /// Maps a user notification to when it was last seen.
    ///
    /// The user notification will not be sent until after `max_age_seconds` from last seen.
    ///
    /// Each time we see a given user notification, the last seen time will be updated.
    items: HashMap<UserNotification, Instant>,
}
impl History {
    pub fn new(max_age_seconds: Duration) -> Self {
        Self {
            max_age_seconds,
            items: HashMap::default(),
        }
    }

    pub fn sendable(&mut self, user_notification: UserNotification) -> Option<UserNotification> {
        let now = Instant::now();
        self.prune(&now); // remove expired items first
        let contains_key = self.items.contains_key(&user_notification);
        self.items.insert(user_notification.clone(), now);
        if contains_key {
            info!("previously sent");
            return None;
        }
        info!("not sent before");
        Some(user_notification)
    }

    pub fn prune(&mut self, now: &Instant) {
        if let Some(threshold) = now.checked_sub(self.max_age_seconds) {
            self.items.retain(|_, last_seen| *last_seen > threshold);
        };
    }
}
