//
use derive_more::derive::Constructor;

pub use history::History;
use kameo::{mailbox::unbounded::UnboundedMailbox, Actor};

use crate::{
    core::{git::UserNotification, server::Shout},
    default_on_actor_link_died, default_on_actor_panic, default_on_actor_start,
    default_on_actor_stop,
};

mod desktop;
mod email;
mod handlers;
mod history;
pub mod messages;
mod webhook;

#[cfg(test)]
mod tests;

#[allow(clippy::module_name_repetitions)]
#[derive(Debug, Constructor)]
pub struct AlertsActor {
    shout: Option<Shout>, // config for sending alerts to users
    history: History,     // record of alerts sent recently (e.g. 24 hours)
    net: kxio::net::Net,
}
impl Actor for AlertsActor {
    type Mailbox = UnboundedMailbox<Self>;
    type Error = crate::IoSendError;

    default_on_actor_start!(this, actor_ref);
    default_on_actor_panic!(this, actor_ref, err);
    default_on_actor_link_died!(this, actor_ref, id, reason);
    default_on_actor_stop!(this, actor_ref, reason);
}

fn short_message(user_notification: &UserNotification) -> String {
    let (forge_alias, repo_alias) = user_notification.aliases();
    format!("[git-next] {forge_alias}/{repo_alias}: {user_notification}")
}

fn full_message(user_notification: &UserNotification) -> String {
    match user_notification {
        UserNotification::CICheckFailed {
            forge_alias,
            repo_alias,
            commit,
            log,
        } => {
            let sha = commit.sha();
            let message = commit.message();
            [
                "CI Checks have Failed".to_string(),
                format!("Forge: {forge_alias}\nRepo : {repo_alias}"),
                format!("Commit:\n - {sha}\n - {message}"),
                "Log:".to_string(),
                log.join("\n"),
            ]
            .join("\n\n")
        }
        UserNotification::RepoConfigLoadFailure {
            forge_alias,
            repo_alias,
            reason,
        } => [
            "Failed to read or parse the .git-next.toml file from repo".to_string(),
            format!(" - {reason}"),
            format!("Forge: {forge_alias}\nRepo : {repo_alias}"),
        ]
        .join("\n\n"),
        UserNotification::WebhookRegistration {
            forge_alias,
            repo_alias,
            reason,
        } => [
            "Failed to register webhook with the forge".to_string(),
            format!(" - {reason}"),
            format!("Forge: {forge_alias}\nRepo : {repo_alias}"),
        ]
        .join("\n\n"),
        UserNotification::DevNotBasedOnMain {
            forge_alias,
            repo_alias,
            dev_branch,
            main_branch,
            dev_commit: _,
            main_commit: _,
            log,
        } => [
            format!("The branch '{dev_branch}' is not based on the branch '{main_branch}'."),
            format!("Action Required: Rebase '{dev_branch}' onto '{main_branch}'."),
            format!("Forge: {forge_alias}\nRepo : {repo_alias}"),
            "Log:".to_string(),
            log.join("\n"),
        ]
        .join("\n\n"),
    }
}
