use std::collections::HashMap;

use bon::Builder;

use crate::core::config::{ApiToken, ForgeType, Hostname, RepoAlias, ServerRepoConfig, User};
use crate::{newtype, s};

use super::CommitCount;

newtype!(
    ServerRepoConfigMap,
    HashMap<RepoAlias, ServerRepoConfig>,
    Default,
    "Maps of repo alias to its config"
);

/// Defines a Forge to connect to
/// Maps from `git-next-server.toml` at `forge.{forge}`
#[derive(
    Builder,
    Clone,
    Debug,
    derive_more::From,
    PartialEq,
    Eq,
    serde::Serialize,
    serde::Deserialize,
    derive_more::Display,
)]
#[display("{}:{}@{}", s!(forge_type).to_lowercase(), user, hostname)]
pub struct ForgeConfig {
    forge_type: ForgeType,
    hostname: String,
    user: String,
    token: String,
    max_dev_commits: Option<u32>,
    pub repos: ServerRepoConfigMap,
}
impl ForgeConfig {
    pub(crate) const fn forge_type(&self) -> ForgeType {
        self.forge_type
    }

    pub(crate) fn hostname(&self) -> Hostname {
        Hostname::new(&self.hostname)
    }

    pub(crate) fn user(&self) -> User {
        User::new(self.user.clone())
    }

    pub(crate) fn token(&self) -> ApiToken {
        ApiToken::new(self.token.clone().into())
    }

    pub(crate) fn max_dev_commits(&self) -> Option<CommitCount> {
        self.max_dev_commits.map(CommitCount::from)
    }

    #[cfg(test)]
    #[must_use]
    pub fn get_repo(&self, repo_alias: impl Into<RepoAlias>) -> Option<&ServerRepoConfig> {
        let repo_alias = repo_alias.into();
        self.repos.get(&repo_alias)
    }
}
