//
use crate::{
    core::{git::repository::Direction, BranchName, ForgeDetails, Hostname, RemoteUrl, RepoPath},
    err, git,
};
use crate::{s, Result};

use derive_more::Constructor;
use gix::bstr::BStr;
use tracing::{info, warn};

use std::{
    borrow::ToOwned,
    ops::Deref,
    path::{Path, PathBuf},
    sync::{Arc, RwLock},
};

use super::{GitExec, OpenRepositoryLike};

#[derive(Clone, Debug, Constructor)]
pub struct RealOpenRepository {
    inner: Arc<RwLock<gix::ThreadSafeRepository>>,
    forge_details: ForgeDetails,
}
impl RealOpenRepository {
    fn git_dir(&self) -> crate::Result<PathBuf> {
        self.inner
            .read()
            .map_err(|_| err!("Lock"))
            .map(|r| r.git_dir().to_path_buf())
    }
}
impl super::OpenRepositoryLike for RealOpenRepository {
    fn remote_branches(&self) -> Result<Vec<BranchName>> {
        let refs = self
            .inner
            .read()
            .map_err(|_| err!("read"))
            .and_then(|repo| {
                repo.to_thread_local()
                    .references()
                    .map_err(|_| err!("thread local references"))
                    .and_then(|refs| {
                        Ok(refs.remote_branches().map(|rb| {
                            rb.filter_map(Result::ok)
                                .map(|r| r.name().to_owned())
                                .map(|n| s!(n))
                                .filter_map(|p| {
                                    p.strip_prefix("refs/remotes/origin/")
                                        .map(ToOwned::to_owned)
                                })
                                .filter(|b| b.as_str() != "HEAD")
                                .map(BranchName::new)
                                .collect::<Vec<_>>()
                        })?)
                    })
            })?;
        Ok(refs)
    }

    #[tracing::instrument]
    fn find_default_remote(&self, direction: git::repository::Direction) -> Option<RemoteUrl> {
        let Ok(repository) = self.inner.read() else {
            #[cfg(not(tarpaulin_include))] // don't test mutex lock failure
            tracing::debug!("no repository");
            return None;
        };
        let thread_local = repository.to_thread_local();
        let Some(Ok(remote)) = thread_local.find_default_remote(direction.into()) else {
            #[cfg(not(tarpaulin_include))] // test is on local repo - should always have remotes
            tracing::debug!("no remote");
            return None;
        };
        remote
            .url(direction.into())
            .cloned()
            .and_then(|url| RemoteUrl::try_from(url).ok())
    }

    #[tracing::instrument(skip_all)]
    #[cfg(not(tarpaulin_include))] // would require writing to external service
    fn fetch(&self) -> Result<()> {
        use crate::{core::git::repository::open::GitExec, s};

        if self
            .find_default_remote(git::repository::Direction::Fetch)
            .is_none()
        {
            return Err(err!("No Fetch Remote Found"));
        }
        info!("Fetching");

        GitExec {
            exec: s!("fetch --prune"),
            git_dir: self.git_dir()?,
        }
        .exec()?;

        info!("Fetch okay");
        Ok(())
    }

    #[cfg(not(tarpaulin_include))] // would require writing to external service
    #[tracing::instrument(skip_all)]
    fn push(
        &self,
        repo_details: &git::RepoDetails,
        branch_name: &BranchName,
        to_commit: &git::GitRef,
        force: &git::push::Force,
    ) -> Result<()> {
        use secrecy::ExposeSecret as _;

        use crate::core::git::{push::Force, repository::open::GitExec};

        let origin = repo_details.origin();
        let force = if let Force::From(old_ref) = force {
            format!("--force-with-lease={branch_name}:{old_ref}")
        } else {
            String::new()
        };
        GitExec {
            exec: format!(
                "push {} {to_commit}:{branch_name} {force}",
                origin.expose_secret()
            ),
            git_dir: self.git_dir()?,
        }
        .exec()?;
        Ok(())
    }

    fn commit_log(
        &self,
        branch_name: &BranchName,
        find_commits: &[git::Commit],
    ) -> Result<Vec<git::Commit>> {
        let limit: usize = if find_commits.is_empty() {
            1
        } else {
            self.forge_details
                .max_dev_commits()
                .map_or(25, |commit_count| commit_count.clone().peel() as usize)
        };
        self.inner.read().map_err(|_| err!("Lock")).map(|repo| {
            let branch = format!("remotes/origin/{branch_name}");
            let branch = BStr::new(&branch);
            let thread_local = repo.to_thread_local();
            let branch_head = thread_local
                .rev_parse_single(branch)
                .map_err(|e| s!(e))
                .map_err(as_gix_error(branch_name.clone()))?;
            let object = branch_head
                .object()
                .map_err(|e| s!(e))
                .map_err(as_gix_error(branch_name.clone()))?;
            let commit = object
                .try_into_commit()
                .map_err(|e| s!(e))
                .map_err(as_gix_error(branch_name.clone()))?;
            let walk = thread_local
                .rev_walk([commit.id])
                .all()
                .map_err(|e| s!(e))
                .map_err(as_gix_error(branch_name.clone()))?;
            let mut commits = vec![];
            for item in walk.take(limit) {
                let item = item
                    .map_err(|e| s!(e))
                    .map_err(as_gix_error(branch_name.clone()))?;
                let commit = item
                    .object()
                    .map_err(|e| s!(e))
                    .map_err(as_gix_error(branch_name.clone()))?;
                let id = commit.id();
                let message = commit
                    .message_raw()
                    .map_err(|e| s!(e))
                    .map_err(as_gix_error(branch_name.clone()))?;
                let commit = git::Commit::new(
                    git::commit::Sha::new(s!(id)),
                    git::commit::Message::new(s!(message)),
                );
                if find_commits.contains(&commit) {
                    commits.push(commit);
                    break;
                }
                commits.push(commit);
            }
            Ok(commits)
        })?
    }

    #[tracing::instrument(skip_all, fields(%branch_name, ?file_name))]
    fn read_file(&self, branch_name: &BranchName, file_name: &Path) -> Result<String> {
        self.inner
            .read()
            .map_err(|_| err!("Lock"))
            .and_then(|repo| {
                let thread_local = repo.to_thread_local();
                let fref = thread_local.find_reference(format!("origin/{branch_name}").as_str())?;
                let id = fref.try_id().ok_or(git::file::Error::TryId)?;
                let oid = id.detach();
                let obj = thread_local.find_object(oid)?;
                let commit = obj.into_commit();
                let tree = commit.tree()?;
                let ent = tree
                    .find_entry(s!(file_name.to_string_lossy()))
                    .ok_or(git::file::Error::FileNotFound)?;
                let fobj = ent.object()?;
                let blob = fobj.into_blob().take_data();
                let content = String::from_utf8(blob)?;
                Ok(content)
            })
    }
}

#[derive(Clone, Debug, Constructor)]
pub struct NonBareRealOpenRepository {
    inner: RealOpenRepository,
}
impl Deref for NonBareRealOpenRepository {
    type Target = RealOpenRepository;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl OpenRepositoryLike for NonBareRealOpenRepository {
    fn remote_branches(&self) -> Result<Vec<BranchName>> {
        self.deref().remote_branches()
    }

    fn find_default_remote(&self, direction: Direction) -> Option<RemoteUrl> {
        self.deref().find_default_remote(direction)
    }

    fn fetch(&self) -> Result<()> {
        self.deref().fetch()
    }

    fn push(
        &self,
        repo_details: &git::RepoDetails,
        branch_name: &BranchName,
        to_commit: &git::GitRef,
        force: &git::push::Force,
    ) -> Result<()> {
        self.deref()
            .push(repo_details, branch_name, to_commit, force)
    }

    fn commit_log(
        &self,
        branch_name: &BranchName,
        find_commits: &[git::Commit],
    ) -> Result<Vec<git::Commit>> {
        self.deref().commit_log(branch_name, find_commits)
    }

    fn read_file(&self, branch_name: &BranchName, file_name: &Path) -> Result<String> {
        self.deref().read_file(branch_name, file_name)
    }
}
impl super::NonBareRepositoryLike for NonBareRealOpenRepository {
    fn checkout(&self, branch_name: &BranchName) -> Result<()> {
        GitExec {
            exec: format!("checkout -b {branch_name}"),
            git_dir: self.git_dir()?,
        }
        .exec()
    }

    fn rebase(&self, target_branch_name: &BranchName) -> Result<()> {
        GitExec {
            exec: format!("rebase {target_branch_name}"),
            git_dir: self.git_dir()?,
        }
        .exec()
    }
}

fn as_gix_error(branch: BranchName) -> impl FnOnce(String) -> color_eyre::eyre::Error {
    move |error| err!("gix: branch: {branch}: {error}")
}

impl From<&RemoteUrl> for git::GitRemote {
    fn from(url: &RemoteUrl) -> Self {
        let path = url.path.clone();
        let path = path.strip_prefix('/').map_or(path.as_str(), |path| path);
        let path = path.strip_suffix(".git").map_or(path, |path| path);
        Self::new(Hostname::new(url.host.clone()), RepoPath::new(s!(path)))
    }
}
