//
use crate::core::{
    git::{self, forge::webhook::Error},
    server::RepoListenUrl,
    RegisteredWebhook, WebhookAuth, WebhookId,
};

use kxio::net::Net;
use secrecy::ExposeSecret as _;
use serde_json::json;
use tracing::{info, instrument, warn};

use crate::forges::forgejo::webhook;
use crate::forges::forgejo::webhook::Hook;

#[instrument(skip_all, fields(forge = %repo_details.forge.forge_alias(), repo = %repo_details.repo_alias))]
pub async fn register(
    repo_details: &git::RepoDetails,
    repo_listen_url: &RepoListenUrl,
    net: &Net,
) -> git::forge::webhook::Result<RegisteredWebhook> {
    let Some(repo_config) = repo_details.repo_config.clone() else {
        return Err(Error::NoRepoConfig);
    };

    // remove any lingering webhooks for the same URL
    let existing_webhook_ids = webhook::list(repo_details, repo_listen_url, net).await?;
    for webhook_id in existing_webhook_ids {
        webhook::unregister(&webhook_id, repo_details, net).await?;
    }

    let hostname = &repo_details.forge.hostname();
    let repo_path = &repo_details.repo_path;
    let token = repo_details.forge.token().expose_secret();
    let url = format!("https://{hostname}/api/v1/repos/{repo_path}/hooks?token={token}");
    let authorisation = WebhookAuth::generate();
    match net
        .post(url)
        .header("Content-Type", "application/json")
        .body(json!({
            "active": true,
            "authorization_header": authorisation.header_value(),
            "branch_filter": format!("{{{},{},{}}}", repo_config.branches().main(), repo_config.branches().next(), repo_config.branches().dev()),
            "config": {
                "content_type": "json",
                "url": repo_listen_url.to_string(),
            },
            "events": [ "push" ],
            "type": "forgejo"
            }).to_string())
        .send()
        .await
    {
        Ok(response) => {
            let Ok(hook) = response.json::<Hook>().await else {
                #[cfg(not(tarpaulin_include))]
                // request response is Json so response_body never returns None
                return Err(Error::NetworkResponseEmpty);
            };
            info!(webhook_id = %hook.id, "ok");
            Ok(RegisteredWebhook::new(
                WebhookId::new(format!("{}", hook.id)),
                authorisation,
            ))
        }
        Err(e) => {
            warn!(?e, "failed");
            Err(Error::FailedToRegister(e.to_string()))
        }
    }
    // Ok(())
}
