//
use color_eyre::Result;
use kameo::message::{Context, Message};
use tracing::warn;

use crate::{
    core::{git, RepoConfigSource},
    repo::{
        branch::advance_main,
        do_send,
        messages::{AdvanceMain, LoadConfigFromRepo, ValidateRepo},
        RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Message<AdvanceMain> for RepoActor {
    type Reply = Result<()>;
    #[tracing::instrument(name = "RepoActor::AdvanceMainTo", skip_all, fields(repo = %self.repo_details, commit = ?msg))]
    async fn handle(
        &mut self,
        msg: AdvanceMain,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        let Some(repo_config) = self.repo_details.repo_config.clone() else {
            warn!("No config loaded");
            return Ok(());
        };
        let Some(open_repository) = &self.open_repository else {
            return Ok(());
        };
        let commit = msg.peel();
        self.post_repo_update(RepoUpdate::AdvancingMain {
            commit: commit.clone(),
        })
        .await?;
        if let Err(err) = advance_main(commit, &self.repo_details, &repo_config, &**open_repository)
        {
            warn!("advance main: {err}");
            self.alert_tui(format!("advance main: {err}")).await?;
        } else {
            self.post_repo_update(RepoUpdate::MainUpdated).await?;
            if let Some(open_repository) = &self.open_repository {
                match open_repository.fetch() {
                    Ok(()) => {
                        self.update_tui_log(git::graph::log(&self.repo_details))
                            .await?;
                    }
                    Err(err) => self.alert_tui(format!("fetching: {err}")).await?,
                }
            }
            match repo_config.source() {
                RepoConfigSource::Repo => {
                    do_send(&ctx.actor_ref(), LoadConfigFromRepo, self.log.as_ref()).await?;
                }
                RepoConfigSource::Server => {
                    do_send(
                        &ctx.actor_ref(),
                        ValidateRepo::new(self.message_token),
                        self.log.as_ref(),
                    )
                    .await?;
                }
            }
        }
        Ok(())
    }
}
