//
use color_eyre::Result;
use kameo::message::{Context, Message};
use tracing::{instrument, warn};

use crate::{
    core::git,
    repo::{
        branch::{advance_next, find_next_commit_on_dev},
        do_send, logger,
        messages::{AdvanceNext, AdvanceNextPayload, ValidateRepo},
        RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Message<AdvanceNext> for RepoActor {
    type Reply = Result<()>;

    #[instrument(skip(self, ctx))]
    async fn handle(
        &mut self,
        msg: AdvanceNext,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        tracing::debug!("start");
        logger(self.log.as_ref(), "start: AdvanceNext").await;
        let Some(repo_config) = &self.repo_details.repo_config else {
            tracing::debug!("repo details.repo_config is unset");
            return Ok(());
        };
        let Some(open_repository) = &self.open_repository else {
            tracing::debug!("no open repository");
            return Ok(());
        };
        tracing::debug!("prerequisites okay");

        let AdvanceNextPayload {
            next,
            main,
            dev_commit_history,
        } = msg.peel();
        tracing::debug!(?next, ?main, ?dev_commit_history, "can we advance?");
        let (commit, force) = find_next_commit_on_dev(&next, &main, &dev_commit_history);
        tracing::debug!(?commit, ?force, "advance?");
        let Some(commit) = commit else {
            tracing::debug!("next is on dev - not advancing next");
            self.alert_tui("Next is on Dev").await?;
            return Ok(());
        };
        tracing::debug!(?commit, "we have a commit to advance to");
        self.post_repo_update(RepoUpdate::AdvancingNext {
            commit: commit.clone(),
            force: force.clone(),
        })
        .await?;
        match advance_next(
            commit,
            force,
            &self.repo_details,
            repo_config,
            &**open_repository,
            self.message_token,
        ) {
            Ok(message_token) => {
                self.post_repo_update(RepoUpdate::NextUpdated).await?;
                match open_repository.fetch() {
                    Ok(()) => {
                        self.update_tui_log(git::graph::log(&self.repo_details))
                            .await?;
                    }
                    Err(err) => self.alert_tui(format!("fetching: {err}")).await?,
                }
                // INFO: pause to allow any CI checks to be started
                tracing::debug!(duration_ms = %self.sleep_duration.as_millis(), "sleeping to allow CI to start");
                tokio::time::sleep(self.sleep_duration).await;
                tracing::debug!("sleeping for CI finished");
                Ok(do_send(
                    &ctx.actor_ref(),
                    ValidateRepo::new(message_token),
                    self.log.as_ref(),
                )
                .await?)
            }
            Err(err) => {
                warn!("advance next: {err}");
                self.post_repo_update(RepoUpdate::AdvanceError(err)).await?;
                Ok(())
            }
        }
    }
}
