//
use color_eyre::Result;
use kameo::message::{Context, Message};
use tracing::debug;

use crate::{
    core::git::{forge::commit::Status, graph, UserNotification},
    repo::{
        do_send, logger,
        messages::{AdvanceMain, ReceiveCIStatus, ValidateRepo},
        notify_user, RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Message<ReceiveCIStatus> for RepoActor {
    type Reply = Result<()>;

    async fn handle(
        &mut self,
        msg: ReceiveCIStatus,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        logger(self.log.as_ref(), "start: ReceiveCIStatus").await;
        let (next, status) = msg.peel();
        self.post_repo_update(RepoUpdate::ReceiveCIStatus {
            status: status.clone(),
        })
        .await?;
        debug!(?status, "");
        let graph_log = graph::log(&self.repo_details);
        self.update_tui_log(graph_log.clone()).await?;
        match status {
            Status::Pass => {
                do_send(&ctx.actor_ref(), AdvanceMain::new(next), self.log.as_ref()).await?;
            }
            Status::Pending => {
                tokio::time::sleep(self.sleep_duration).await;
                do_send(
                    &ctx.actor_ref(),
                    ValidateRepo::new(self.message_token),
                    self.log.as_ref(),
                )
                .await?;
            }
            Status::Fail => {
                tracing::warn!("Checks have failed");
                notify_user(
                    self.notify_user_recipient.as_ref(),
                    UserNotification::CICheckFailed {
                        forge_alias: self.repo_details.forge.forge_alias().clone(),
                        repo_alias: self.repo_details.repo_alias.clone(),
                        commit: next,
                        log: graph_log,
                    },
                    self.log.as_ref(),
                )
                .await?;
                debug!("sleeping before retrying...");
                logger(self.log.clone().as_ref(), "before sleep").await;
                tokio::time::sleep(self.sleep_duration).await;
                logger(self.log.clone().as_ref(), "after sleep").await;
                do_send(
                    &ctx.actor_ref(),
                    ValidateRepo::new(self.message_token),
                    self.log.as_ref(),
                )
                .await?;
            }
            Status::Error(err) => {
                tracing::warn!(?err, "Check CI Status");
            }
        }
        Ok(())
    }
}
