//
use std::collections::HashMap;

use kameo::message::{Context, Message};
use tracing::{debug, info, instrument, warn};

use crate::{
    core::git::{
        push::Force,
        validation::positions::{validate, Positions, PositionsError},
        UserNotification,
    },
    repo::{
        do_send, logger,
        messages::{AdvanceNext, AdvanceNextPayload, CheckCIStatus, MessageToken, ValidateRepo},
        notify_user, RepoActor,
    },
    s,
    server::actor::messages::RepoUpdate,
    Result,
};

impl Message<ValidateRepo> for RepoActor {
    type Reply = Result<()>;

    #[instrument(name = "RepoActor::ValidateRepo", skip_all, fields(repo = %self.repo_details))]
    async fn handle(
        &mut self,
        msg: ValidateRepo,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        logger(self.log.as_ref(), "start: ValidateRepo").await;
        // Message Token - make sure we are only triggered for the latest/current token
        match self.token_status(msg.peel()) {
            TokenStatus::Current => {} // do nothing
            TokenStatus::Expired => {
                logger(
                    self.log.as_ref(),
                    format!("discarded: old message token: {}", self.message_token),
                )
                .await;
                self.post_repo_update(RepoUpdate::Waiting).await?;
                return Ok(()); // message is expired
            }
            TokenStatus::New(message_token) => {
                self.message_token = message_token;
                self.post_repo_update(RepoUpdate::UpdateMessageToken(self.message_token))
                    .await?;
                logger(
                    self.log.as_ref(),
                    format!("new message token: {}", self.message_token),
                )
                .await;
            }
        }
        logger(
            self.log.as_ref(),
            format!("accepted token: {}", self.message_token),
        )
        .await;
        self.post_repo_update(RepoUpdate::ValidateRepo).await?;
        // Repository positions
        let Some(ref open_repository) = self.open_repository else {
            logger(self.log.as_ref(), "no open repository").await;
            self.alert_tui("repo not open").await?;
            self.post_repo_update(RepoUpdate::NotOpen).await?;
            return Ok(());
        };
        logger(self.log.as_ref(), "have open repository").await;
        let Some(repo_config) = self.repo_details.repo_config.clone() else {
            logger(self.log.as_ref(), "no repo config").await;
            self.alert_tui("no repo config").await?;
            self.post_repo_update(RepoUpdate::NoConfig).await?;
            return Ok(());
        };
        logger(self.log.as_ref(), "have repo config").await;
        info!("validating positions");
        match validate(&**open_repository, &self.repo_details, &repo_config) {
            Ok((
                Positions {
                    main,
                    next,
                    dev,
                    dev_commit_history,
                    next_is_valid,
                },
                git_log,
            )) => {
                info!("positions - ok");
                self.post_repo_update(RepoUpdate::Okay).await?;
                self.last_main_commit = Some(main.clone());
                self.last_next_commit = Some(next.clone());
                self.last_dev_commit = Some(dev.clone());
                let mut positions = HashMap::new();
                positions
                    .entry(s!(s!(main.sha())[0..7]))
                    .or_insert_with(Vec::new)
                    .extend(["main"]);
                positions
                    .entry(s!(s!(next.sha())[0..7]))
                    .or_insert_with(Vec::new)
                    .extend(["next"]);
                positions
                    .entry(s!(s!(dev.sha())[0..7]))
                    .or_insert_with(Vec::new)
                    .extend(["dev"]);
                info!(?positions, "");
                self.update_tui_log(git_log).await?;
                if next_is_valid && next != main {
                    info!("Checking CI");
                    do_send(
                        &ctx.actor_ref(),
                        CheckCIStatus::new(next),
                        self.log.as_ref(),
                    )
                    .await?;
                } else if next != dev {
                    info!("Advance next");
                    self.post_repo_update(RepoUpdate::AdvancingNext {
                        commit: next.clone(),
                        force: Force::No,
                    })
                    .await?;
                    do_send(
                        &ctx.actor_ref(),
                        AdvanceNext::new(AdvanceNextPayload {
                            next,
                            main,
                            dev_commit_history,
                        }),
                        self.log.as_ref(),
                    )
                    .await?;
                    info!("advance next sent");
                } else {
                    info!("do nothing");
                    self.post_repo_update(RepoUpdate::Okay).await?;
                }
            }
            Err(PositionsError::Retryable(message)) => {
                warn!(?message, "Retryable");
                self.alert_tui(format!("retryable: {message}")).await?;
                logger(self.log.as_ref(), message).await;
                debug!("sleeping before retrying...");
                tokio::time::sleep(self.sleep_duration).await;
                do_send(
                    &ctx.actor_ref(),
                    ValidateRepo::new(self.message_token),
                    self.log.as_ref(),
                )
                .await?;
            }
            Err(PositionsError::UserIntervention(user_notification)) => {
                warn!(?user_notification, "User Intervention");
                self.alert_tui(format!("USER INTERVENTION: {user_notification}"))
                    .await?;
                self.post_repo_update(RepoUpdate::UserIntervention(user_notification.clone()))
                    .await?;
                if let UserNotification::DevNotBasedOnMain {
                    main_commit,
                    dev_commit,
                    ..
                } = &user_notification
                {
                    self.last_main_commit = Some(main_commit.clone());
                    self.last_dev_commit = Some(dev_commit.clone());
                }
                if let UserNotification::CICheckFailed { log, .. }
                | UserNotification::DevNotBasedOnMain { log, .. } = &user_notification
                {
                    self.update_tui_log(log.clone()).await?;
                }
                notify_user(
                    self.notify_user_recipient.as_ref(),
                    user_notification,
                    self.log.as_ref(),
                )
                .await?;
            }
            Err(PositionsError::NonRetryable(message)) => {
                warn!(?message, "NonRetryable");
                self.alert_tui(format!("Error: {message}")).await?;
                self.post_repo_update(RepoUpdate::Error(message.clone()))
                    .await?;
                logger(self.log.as_ref(), message).await;
            }
        }
        Ok(())
    }
}

enum TokenStatus {
    Current,
    Expired,
    New(MessageToken),
}
impl RepoActor {
    fn token_status(&self, new: MessageToken) -> TokenStatus {
        let current = &self.message_token;
        if &new > current {
            return TokenStatus::New(new);
        }
        if current > &new {
            return TokenStatus::Expired;
        }
        TokenStatus::Current
    }
}
