//
use color_eyre::Result;
use kameo::message::{Context, Message};
use tracing::{info, instrument, warn};

use crate::{
    core::{
        git::{Commit, ForgeLike},
        webhook::{push::Branch, Push},
        BranchName, WebhookAuth,
    },
    repo::{
        do_send, logger,
        messages::{ValidateRepo, WebhookNotification},
        ActorLog, RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Message<WebhookNotification> for RepoActor {
    type Reply = Result<()>;

    #[instrument(name = "RepoActor::WebhookMessage", skip_all, fields(token = %self.message_token, repo = %self.repo_details))]
    async fn handle(
        &mut self,
        msg: WebhookNotification,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        let Some(config) = &self.repo_details.repo_config else {
            logger(self.log.as_ref(), "server has no repo config").await;
            warn!("No repo config");
            return Ok(());
        };
        if validate_notification(
            &msg,
            self.webhook_auth.as_ref(),
            &*self.forge,
            self.log.as_ref(),
        )
        .await
        .is_err()
        {
            return Ok(());
        }
        let push = match self.forge.parse_webhook_body(msg.body()) {
            Err(err) => {
                logger(self.log.as_ref(), "message parse error - not a push").await;
                warn!(?err, "Not a 'push'");
                return Ok(());
            }
            Ok(push) => match push.branch(config.branches()) {
                None => {
                    logger(self.log.as_ref(), "unknown branch").await;
                    warn!(
                        ?push,
                        "Unrecognised branch, we should be filtering to only the ones we want"
                    );
                    return Ok(());
                }
                Some(Branch::Main) => {
                    handle_push(
                        push.clone(),
                        &config.branches().main(),
                        &mut self.last_main_commit,
                        self.log.as_ref(),
                    )
                    .await;
                    push
                }
                Some(Branch::Next) => {
                    handle_push(
                        push.clone(),
                        &config.branches().next(),
                        &mut self.last_next_commit,
                        self.log.as_ref(),
                    )
                    .await;
                    push
                }
                Some(Branch::Dev) => {
                    handle_push(
                        push.clone(),
                        &config.branches().dev(),
                        &mut self.last_dev_commit,
                        self.log.as_ref(),
                    )
                    .await;
                    push
                }
            },
        };
        self.post_repo_update(RepoUpdate::WebhookReceived { push: push.clone() })
            .await?;
        let message_token = self.message_token.next();
        info!(
            token = %message_token,
            "New commit"
        );
        do_send(
            &ctx.actor_ref(),
            ValidateRepo::new(message_token),
            self.log.as_ref(),
        )
        .await?;
        Ok(())
    }
}

async fn validate_notification(
    msg: &WebhookNotification,
    webhook_auth: Option<&WebhookAuth>,
    forge: &dyn ForgeLike,
    log: Option<&ActorLog>,
) -> Result<(), ()> {
    let Some(expected_authorization) = webhook_auth else {
        logger(log, "server has no auth token").await;
        warn!("Don't know what authorization to expect");
        return Err(());
    };

    if !forge.is_message_authorised(msg, expected_authorization) {
        logger(log, "message authorisation is invalid").await;
        warn!(
            "Invalid authorization - expected {}",
            expected_authorization
        );
        return Err(());
    }
    if forge.should_ignore_message(msg) {
        logger(log, "forge sent ignorable message").await;
        return Err(());
    }
    Ok(())
}

async fn handle_push(
    push: Push,
    branch: &BranchName,
    last_commit: &mut Option<Commit>,
    log: Option<&ActorLog>,
) {
    logger(log, format!("message is for {branch} branch")).await;
    let commit = Commit::from(push);
    if last_commit.as_ref() == Some(&commit) {
        logger(log, format!("not a new commit on {branch}")).await;
        info!(
            %branch ,
            %commit,
            "Ignoring - already aware of branch at commit",
        );
    }
    last_commit.replace(commit);
}
