//
use std::sync::Arc;

use color_eyre::{eyre::eyre, Result};
use derive_more::with_trait::Deref;
use kameo::{actor::ActorRef, mailbox::unbounded::UnboundedMailbox, Actor};
use kxio::net::Net;
use tokio::sync::RwLock;
use tracing::{debug, info, instrument, warn};

use crate::{
    alerts::{messages::NotifyUser, AlertsActor},
    core::{
        git::{
            self,
            repository::{factory::RepositoryFactory, open::OpenRepositoryLike},
            UserNotification,
        },
        server::ListenUrl,
        WebhookAuth, WebhookId,
    },
    default_on_actor_link_died, default_on_actor_panic, default_on_actor_start, on_actor_stop,
    server::{
        actor::messages::{RepoUpdate, ServerUpdate},
        server_mode::ServerMode,
        ServerActor,
    },
    tell, IoSendError,
};

pub use branch::AdvanceError;

mod branch;
pub mod handlers;
mod load;
pub mod messages;
mod notifications;

#[cfg(test)]
pub mod tests;

#[derive(Clone, Debug, Default)]
pub struct ActorLog(Arc<RwLock<Vec<String>>>);
impl Deref for ActorLog {
    type Target = Arc<RwLock<Vec<String>>>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
impl ActorLog {
    pub async fn log(&self, message: impl Into<String> + Send) {
        let message = message.into();
        debug!(%message, "log");
        self.write().await.push(message);
    }
}

/// An actor that represents a Git Repository.
///
/// When this actor is started it is sent the `CloneRepo` message.
#[allow(clippy::module_name_repetitions)]
#[derive(Debug, derive_more::Display, derive_with::With)]
#[display("{}:{}:{}", generation, repo_details.forge.forge_alias(), repo_details.repo_alias)]
pub struct RepoActor {
    server_mode: ServerMode,
    sleep_duration: std::time::Duration,
    generation: git::Generation,
    message_token: messages::MessageToken,
    repo_details: git::RepoDetails,
    listen_url: ListenUrl,
    webhook_id: Option<WebhookId>, // INFO: if [None] then no webhook is configured
    webhook_auth: Option<WebhookAuth>, // INFO: if [None] then no webhook is configured
    last_main_commit: Option<git::Commit>,
    last_next_commit: Option<git::Commit>,
    last_dev_commit: Option<git::Commit>,
    repository_factory: Box<dyn RepositoryFactory>,
    open_repository: Option<Box<dyn OpenRepositoryLike>>,
    net: Net,
    forge: Box<dyn git::ForgeLike>,
    log: Option<ActorLog>,
    notify_user_recipient: Option<ActorRef<AlertsActor>>,
    server_actor_ref: ActorRef<ServerActor>,
}
impl RepoActor {
    #[allow(clippy::too_many_arguments)]
    pub fn new(
        server_mode: ServerMode,
        repo_details: git::RepoDetails,
        forge: Box<dyn git::ForgeLike>,
        listen_url: ListenUrl,
        generation: git::Generation,
        net: Net,
        repository_factory: Box<dyn RepositoryFactory>,
        sleep_duration: std::time::Duration,
        notify_user_recipient: Option<ActorRef<AlertsActor>>,
        server_actor_ref: ActorRef<ServerActor>,
    ) -> Self {
        let message_token = messages::MessageToken::default();
        Self {
            server_mode,
            generation,
            message_token,
            repo_details,
            listen_url,
            webhook_id: None,
            webhook_auth: None,
            last_main_commit: None,
            last_next_commit: None,
            last_dev_commit: None,
            repository_factory,
            open_repository: None,
            forge,
            net,
            sleep_duration,
            log: None,
            notify_user_recipient,
            server_actor_ref,
        }
    }

    async fn update_tui_branches(&self) -> Result<()> {
        use crate::server::actor::messages::RepoUpdate;
        let Some(repo_config) = &self.repo_details.repo_config else {
            return Ok(());
        };
        let branches = repo_config.branches().clone();
        self.post_repo_update(RepoUpdate::Branches { branches })
            .await?;
        Ok(())
    }

    #[allow(unused_variables)]
    async fn update_tui_log(&self, log: git::graph::Log) -> Result<()> {
        self.post_repo_update(RepoUpdate::Log { log }).await?;
        Ok(())
    }

    #[allow(unused_variables)]
    async fn alert_tui(&self, alert: impl Into<String> + Send) -> Result<()> {
        self.post_repo_update(RepoUpdate::Alert {
            alert: alert.into(),
        })
        .await?;
        Ok(())
    }

    #[allow(unused_variables)]
    #[instrument(skip_all)]
    async fn post_repo_update(&self, repo_update: RepoUpdate) -> Result<()> {
        match self.server_mode {
            ServerMode::LegacyUI | ServerMode::NewUI => {
                tell!(
                    "server",
                    self.server_actor_ref,
                    ServerUpdate::RepoUpdate {
                        forge_alias: self.repo_details.forge.forge_alias().clone(),
                        repo_alias: self.repo_details.repo_alias.clone(),
                        repo_update,
                    }
                )?;
            }
            ServerMode::Simple => (),
        }
        Ok(())
    }
}
impl Actor for RepoActor {
    type Mailbox = UnboundedMailbox<Self>;
    type Error = IoSendError;

    default_on_actor_start!(this, actor_ref);
    default_on_actor_panic!(this, actor_ref, err);
    default_on_actor_link_died!(this, actor_ref, id, reason);

    on_actor_stop!(this, actor_ref, reason, {
        info!("Checking webhook");
        if let Some(webhook_id) = this.webhook_id.take() {
            tracing::warn!("stopping - unregistering webhook");
            info!(%webhook_id, "Unregistring webhook");
            let forge = this.forge.duplicate();
            if let Err(err) = forge.unregister_webhook(&webhook_id).await {
                warn!("unregistering webhook: {err}");
            }
        }
        Ok(())
    });
}

pub async fn do_send<M>(
    repo_actor_ref: &ActorRef<RepoActor>,
    msg: M,
    log: Option<&ActorLog>,
) -> Result<()>
where
    M: Send + Sync + 'static + std::fmt::Debug,
    RepoActor: kameo::message::Message<M>,
    <RepoActor as kameo::message::Message<M>>::Reply: Send + Sync + 'static + std::fmt::Debug,
    <<RepoActor as kameo::message::Message<M>>::Reply as kameo::Reply>::Error: std::fmt::Debug,
{
    let log_message = format!("send: {msg:?}");
    logger(log, log_message.clone()).await;
    if cfg!(not(test)) {
        tell!(repo_actor_ref, msg).map_err(|e| eyre!(format!("error: {log_message}: {e:?}")))?;
    }
    Ok(())
}

pub async fn logger(log: Option<&ActorLog>, message: impl Into<String> + Send) {
    if let Some(log) = log {
        log.log(message).await;
    }
}
pub async fn notify_user(
    recipient: Option<&ActorRef<AlertsActor>>,
    user_notification: UserNotification,
    log: Option<&ActorLog>,
) -> Result<()> {
    let msg = NotifyUser::from(user_notification);
    let log_message = format!("send: {msg:?}");
    debug!(log_message);
    logger(log, log_message).await;
    if let Some(recipient) = &recipient {
        tell!("alerts", recipient, msg)?;
    }
    Ok(())
}
