//
use color_eyre::Result;
use kameo::{
    actor::ActorRef,
    message::{Context, Message},
};

use crate::{
    alerts::messages::UpdateShout,
    base_actor::BaseActor as _,
    core::{ForgeAlias, RepoAlias},
    repo::{messages::CloneRepo, RepoActor},
    server::actor::{
        messages::{ReceiveValidAppConfig, ServerUpdate, ValidAppConfig},
        ServerActor,
    },
    spawn, tell,
    webhook::{self, router::AddWebhookRecipient, WebhookActor},
};

impl Message<ReceiveValidAppConfig> for ServerActor {
    type Reply = Result<()>;

    async fn handle(
        &mut self,
        msg: ReceiveValidAppConfig,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        let ValidAppConfig {
            app_config,
            socket_address,
            storage: server_storage,
        } = msg.peel();
        // shutdown any existing webhook actor
        if let Some(webhook_actor_ref) = self.webhook_actor_ref.take() {
            webhook_actor_ref.kill();
        }
        self.generation.inc();

        // Webhook Server
        if let Some(webhook_router_actor_ref) = &self.webhook_router_actor_ref {
            let listen_url = app_config.listen().url();
            let server_actor_ref = ctx.actor_ref();
            // Forge Actors
            for (forge_alias, forge_config) in app_config.forges() {
                let repo_actors = self
                    .create_forge_repos(
                        forge_config,
                        forge_alias.clone(),
                        &server_storage,
                        listen_url,
                        &self.alerts,
                        server_actor_ref.clone(),
                    )
                    .into_iter()
                    .map(|repo_actor_tuple| {
                        start_repo_actor(repo_actor_tuple, server_actor_ref.clone())
                    })
                    .collect::<Vec<_>>();
                for repo_actor in repo_actors {
                    let (repo_alias, repo_actor_ref) = repo_actor.await?;
                    tell!(
                        webhook_router_actor_ref,
                        AddWebhookRecipient::new(
                            forge_alias.clone(),
                            repo_alias.clone(),
                            repo_actor_ref.clone(),
                        )
                    )?;

                    self.repo_actors
                        .insert((forge_alias.clone(), repo_alias), repo_actor_ref);
                }
            }
            let webhook_actor_ref = spawn!(
                ctx.actor_ref(),
                WebhookActor::new(socket_address, webhook_router_actor_ref.clone())
            );
            tell!(webhook_actor_ref, webhook::Start)?;
            self.webhook_actor_ref.replace(webhook_actor_ref);
        }
        let shout = app_config.shout().clone();
        self.app_config.replace(app_config.clone());
        tell!(
            "server",
            ctx.actor_ref(),
            ServerUpdate::AppConfigLoaded {
                app_config: ValidAppConfig {
                    app_config,
                    socket_address,
                    storage: server_storage,
                },
            }
        )?;
        tell!("alert", self.alerts, UpdateShout::new(shout))?;
        Ok(())
    }
}

async fn start_repo_actor(
    actor: (ForgeAlias, RepoAlias, RepoActor),
    server_actor_ref: ActorRef<ServerActor>,
) -> Result<(RepoAlias, ActorRef<RepoActor>)> {
    let (forge_name, repo_alias, repo_actor) = actor;
    let span = tracing::info_span!("start_repo_actor", forge = %forge_name, repo = %repo_alias);
    let _guard = span.enter();
    let repo_actor_ref = repo_actor.spawn(server_actor_ref).await;
    tell!(repo_actor_ref, CloneRepo)?;
    Ok((repo_alias, repo_actor_ref))
}
