//
use std::net::SocketAddr;

use derive_more::Constructor;
use kameo::actor::ActorRef;
use tokio::sync::mpsc::Sender;

use crate::message;
use crate::repo::messages::MessageToken;
use crate::repo::AdvanceError;
use crate::{
    core::{
        git::{self, forge::commit::Status, graph::Log, UserNotification},
        server::{AppConfig, Storage},
        webhook::Push,
        ForgeAlias, RepoAlias, RepoBranches, RepoConfig,
    },
    repo::RepoActor,
};

// receive server config
message!(
    ReceiveAppConfig,
    AppConfig,
    "Notification of newly loaded server configuration.

This message will prompt the `git-next` server to stop and restart all repo-actors.

Contains the new server configuration."
);

// receive valid server config
#[derive(Clone, Debug, PartialEq, Eq, Constructor, serde::Serialize, serde::Deserialize)]
pub struct ValidAppConfig {
    pub app_config: AppConfig,
    pub socket_address: SocketAddr,
    pub storage: Storage,
}
message!(
    ReceiveValidAppConfig,
    ValidAppConfig,
    "Notification of validated server configuration."
);

message!(Shutdown, "Notification to shutdown the server actor");

#[allow(clippy::large_enum_variant)]
#[derive(Clone, Debug)]
pub enum ServerUpdate {
    /// List of all configured forges and aliases
    AppConfigLoaded { app_config: ValidAppConfig },

    RepoUpdate {
        forge_alias: ForgeAlias,
        repo_alias: RepoAlias,
        repo_update: RepoUpdate,
    },
}

#[derive(Clone, Debug)]
pub enum RepoUpdate {
    /// Details the configured branches as loaded from the repo's .git-next.toml file
    Branches {
        branches: RepoBranches,
    },
    /// An update of the git branch graph log
    Log {
        log: Log,
    },
    ValidateRepo,
    Okay,
    Alert {
        alert: String,
    },
    CheckingCI,
    AdvancingNext {
        commit: git::Commit,
        force: git::push::Force,
    },
    AdvancingMain {
        commit: git::Commit,
    },
    Opening {
        repo_actor: ActorRef<RepoActor>,
    },
    LoadingConfigFromRepo,
    ReceiveCIStatus {
        status: Status,
    },
    ReceiveRepoConfig {
        repo_config: RepoConfig,
    },
    RegisteringWebhook,
    UnregisteringWebhook,
    WebhookReceived {
        push: Push,
    },
    RegisteredWebhook,
    Opened,
    NextUpdated,
    MainUpdated,
    NotOpen,
    NoConfig,
    UserIntervention(UserNotification),
    Error(String),
    UpdateMessageToken(MessageToken),
    RebaseError(String),
    ProgressMessage(String),
    RebaseCompleted,
    AdvanceError(AdvanceError),
    Waiting,
}

/// Sends a channel to be used to shutdown the server
#[derive(Constructor)]
pub struct ShutdownTrigger(Sender<String>);
impl ShutdownTrigger {
    pub fn peel(self) -> Sender<String> {
        self.0
    }
}
