//
use std::{
    collections::BTreeMap,
    ops::ControlFlow,
    path::PathBuf,
    sync::{Arc, RwLock},
};

use kameo::{
    actor::ActorRef, error::ActorStopReason, mailbox::unbounded::UnboundedMailbox,
    message::Context, Actor,
};
use kxio::{fs::FileSystem, net::Net};
use tokio::sync::mpsc::Sender;
use tracing::{error, instrument, warn};

use crate::{
    alerts::AlertsActor,
    core::{
        git::{repository::factory::RepositoryFactory, Generation, RepoDetails},
        server::{AppConfig, ListenUrl, Storage},
        ForgeAlias, ForgeConfig, GitDir, RepoAlias, ServerRepoConfig, StoragePathType,
    },
    err,
    forge::Forge,
    on_actor_link_died, on_actor_panic, on_actor_start, on_actor_stop,
    repo::RepoActor,
    s, spawn, tell,
    webhook::{router::WebhookRouterActor, WebhookActor},
    IoSendError, MessageBus, Result,
};

use messages::{ReceiveAppConfig, ServerUpdate, Shutdown};

use super::server_mode::ServerMode;

#[cfg(test)]
mod tests;

mod handlers;
pub mod messages;

#[allow(clippy::module_name_repetitions)]
#[derive(derive_with::With)]
#[with(message_log)]
pub struct ServerActor {
    server_mode: ServerMode,
    app_config: Option<AppConfig>,
    generation: Generation,
    webhook_actor_ref: Option<ActorRef<WebhookActor>>,
    fs: FileSystem,
    net: Net,
    alerts: ActorRef<AlertsActor>,
    repository_factory: Box<dyn RepositoryFactory>,
    sleep_duration: std::time::Duration,
    repo_actors: BTreeMap<(ForgeAlias, RepoAlias), ActorRef<RepoActor>>,

    shutdown_trigger: Option<Sender<String>>,
    server_updates_bus: MessageBus<ServerUpdate>,
    webhook_router_actor_ref: Option<ActorRef<WebhookRouterActor>>,

    // testing
    message_log: Option<Arc<RwLock<Vec<String>>>>,
}

impl Actor for ServerActor {
    type Mailbox = UnboundedMailbox<Self>;
    type Error = IoSendError;

    on_actor_start!(this, actor_ref, {
        let webhook_router = spawn!(actor_ref, WebhookRouterActor::default());
        this.webhook_router_actor_ref.replace(webhook_router);
        Ok(())
    });

    on_actor_panic!(this, actor_ref, err, {
        warn!(%err, "paniced: {}", <Self as Actor>::name());
        Ok(ControlFlow::Break(ActorStopReason::Panicked(err)))
    });

    on_actor_link_died!(this, actor_ref, id, reason, {
        match &reason {
            ActorStopReason::Killed | ActorStopReason::Normal => {
                Ok(ControlFlow::Break(ActorStopReason::Normal))
            }
            ActorStopReason::Panicked(_) | ActorStopReason::LinkDied { .. } => {
                if let Some(actor_ref) = actor_ref.upgrade() {
                    tell!(actor_ref, Shutdown)?;
                }
                if let Some(trigger) = this.shutdown_trigger.take() {
                    trigger
                        .send(s!("link died"))
                        .await
                        .map_err(|e| format!("failed sending shutdown trigger: {e:?}"))?;
                }
                Ok(ControlFlow::Break(ActorStopReason::LinkDied {
                    id,
                    reason: Box::new(reason),
                }))
            }
        }
    });

    on_actor_stop!(this, actor_ref, reason, Ok(()));
}

impl ServerActor {
    pub fn new(
        server_mode: ServerMode,
        fs: FileSystem,
        net: Net,
        alerts: ActorRef<AlertsActor>,
        server_updates_bus: MessageBus<ServerUpdate>,
        repo: Box<dyn RepositoryFactory>,
        sleep_duration: std::time::Duration,
    ) -> Self {
        Self {
            server_mode,
            app_config: None,
            generation: Generation::default(),
            webhook_actor_ref: None,
            fs,
            net,
            alerts,
            repository_factory: repo,
            shutdown_trigger: None,
            server_updates_bus,
            webhook_router_actor_ref: None,
            sleep_duration,
            repo_actors: BTreeMap::new(),
            message_log: None,
        }
    }

    fn create_forge_data_directories(
        &self,
        app_config: &AppConfig,
        server_dir: &std::path::Path,
    ) -> Result<()> {
        for (forge_name, _forge_config) in app_config.forges() {
            let path = server_dir.join(PathBuf::from(&forge_name));
            let path_handle = self.fs.path(&path);
            if path_handle.exists()? {
                if !path_handle.is_dir()? {
                    return Err(err!("ForgeDirIsNotDirectory {}", path.display()));
                }
            } else {
                tracing::info!(%forge_name, ?path_handle, "creating storage");
                self.fs.dir(&path).create_all()?;
            }
        }
        Ok(())
    }

    #[instrument(skip_all)]
    fn create_forge_repos(
        &self,
        forge_config: &ForgeConfig,
        forge_name: ForgeAlias,
        server_storage: &Storage,
        listen_url: &ListenUrl,
        notify_user_recipient: &ActorRef<AlertsActor>,
        server_actor_ref: ActorRef<Self>,
    ) -> Vec<(ForgeAlias, RepoAlias, RepoActor)> {
        tracing::info!(%forge_name, %forge_config, "");
        let creator = self.create_actor(
            forge_name,
            forge_config.clone(),
            server_storage,
            listen_url,
            server_actor_ref,
        );
        forge_config
            .repos
            .iter()
            .map(|(repo_alias, server_repo_config)| {
                creator((
                    repo_alias.clone(),
                    server_repo_config,
                    notify_user_recipient.clone(),
                ))
            })
            .collect::<Vec<_>>()
    }

    fn create_actor(
        &self,
        forge_name: ForgeAlias,
        forge_config: ForgeConfig,
        server_storage: &Storage,
        listen_url: &ListenUrl,
        server_actor_ref: ActorRef<Self>,
    ) -> impl Fn(
        (RepoAlias, &ServerRepoConfig, ActorRef<AlertsActor>),
    ) -> (ForgeAlias, RepoAlias, RepoActor) {
        let server_storage = server_storage.clone();
        let listen_url = listen_url.clone();
        let net = self.net.clone();
        let repository_factory = self.repository_factory.duplicate();
        let generation = self.generation;
        let sleep_duration = self.sleep_duration;
        let server_mode = self.server_mode;
        move |(repo_alias, server_repo_config, notify_user_recipient)| {
            let span = tracing::info_span!("create_actor", alias = %repo_alias, config = %server_repo_config);
            let _guard = span.enter();
            let gitdir = server_repo_config.gitdir().unwrap_or_else(|| {
                GitDir::new(
                    server_storage
                        .path()
                        .join(forge_name.to_string())
                        .join(repo_alias.to_string()),
                    StoragePathType::Internal,
                )
            });
            // INFO: can't canonicalise gitdir as the path needs to exist to do that and we may not
            //       have cloned the repo yet
            let repo_details = RepoDetails::new(
                generation,
                &repo_alias,
                server_repo_config,
                &forge_name,
                &forge_config,
                gitdir,
            );
            let forge = Forge::create(repo_details.clone(), net.clone());
            let actor = RepoActor::new(
                server_mode,
                repo_details,
                forge,
                listen_url.clone(),
                generation,
                net.clone(),
                repository_factory.duplicate(),
                sleep_duration,
                Some(notify_user_recipient),
                server_actor_ref.clone(),
            );
            (forge_name.clone(), repo_alias, actor)
        }
    }

    fn server_storage(&self, app_config: &ReceiveAppConfig) -> Option<Storage> {
        let server_storage = app_config.storage().clone();
        let dir = server_storage.path();
        if !dir.exists() {
            if let Err(err) = self.fs.dir(dir).create() {
                error!(?err, ?dir, "Failed to create server storage");
                return None;
            }
        }
        let Ok(canon) = dir.canonicalize() else {
            error!(?dir, "Failed to confirm server storage");
            return None;
        };
        if let Err(err) = self.create_forge_data_directories(app_config, &canon) {
            error!(?err, "Failure creating forge storage");
            return None;
        }
        Some(server_storage)
    }

    /// Attempts to gracefully shutdown the server before stopping the system.
    async fn abort(
        &mut self,
        ctx: &Context<Self, Result<()>>,
        message: impl Into<String> + Send,
    ) -> Result<()> {
        let message = message.into();
        error!(%message, "aborting");
        self.do_send(Shutdown, ctx).await?;
        if let Some(t) = self.shutdown_trigger.take() {
            t.send(message)
                .await
                .map_err(|e| err!("failed sending shutdown trigger: {e:?}"))?;
        }
        Ok(())
    }

    async fn do_send<M>(&self, msg: M, ctx: &Context<Self, Result<()>>) -> Result<()>
    where
        M: Send + Sync + 'static + std::fmt::Debug,
        Self: kameo::message::Message<M>,
        <Self as kameo::message::Message<M>>::Reply: Send + Sync + 'static + std::fmt::Debug,
        <<Self as kameo::message::Message<M>>::Reply as kameo::Reply>::Error: std::fmt::Debug,
    {
        let log_message = format!("send: {msg:?}");
        if let Some(message_log) = &self.message_log {
            if let Ok(mut log) = message_log.write() {
                log.push(log_message.clone());
            }
        }
        if cfg!(not(test)) {
            tell!(ctx.actor_ref(), msg)
                .map_err(|e| err!("failed sending: {log_message}: {e:?}"))?;
        }
        Ok(())
    }
}
