//
use std::{path::PathBuf, sync::Arc};

use color_eyre::{eyre::Context, Result};
use kxio::{fs::FileSystem, net::Net};
use root::RootActor;
use server_mode::ServerMode;
use tokio::sync::RwLock;
use tracing::info;

use crate::core::git::RepositoryFactory;

use crate::tell;

pub use actor::ServerActor;

pub mod actor;
pub mod root;
pub mod server_mode;

#[cfg(test)]
mod tests;

pub fn init(fs: &FileSystem) -> Result<()> {
    let file_name = "git-next-server.toml";
    let pathbuf = PathBuf::from(file_name);
    if fs
        .path(&pathbuf)
        .exists()
        .with_context(|| format!("Checking for existing file: {pathbuf:?}"))?
    {
        eprintln!("The configuration file already exists at {pathbuf:?} - not overwritting it.",);
    } else {
        fs.file(&pathbuf)
            .write(include_str!("server-default.toml"))
            .with_context(|| format!("Writing file: {pathbuf:?}"))?;
        println!("Created a default configuration file at {pathbuf:?}",);
    }
    Ok(())
}

#[allow(clippy::too_many_lines)]
pub fn start(
    server_mode: ServerMode,
    fs: &FileSystem,
    net: &Net,
    repo: Box<dyn RepositoryFactory>,
    sleep_duration: std::time::Duration,
) -> Result<()> {
    match server_mode {
        ServerMode::Simple => init_logging(),
        ServerMode::LegacyUI | ServerMode::NewUI => crate::tui::logging::initialize_logging()?,
    };

    let shutdown_message_holder: Arc<RwLock<Option<String>>> = Arc::new(RwLock::new(None));
    let shutdown_message_holder_clone = shutdown_message_holder.clone();

    #[allow(clippy::expect_used)]
    tokio::runtime::Runtime::new()?.block_on(async {
        let (tx_shutdown, mut rx_shutdown) = tokio::sync::mpsc::channel::<String>(1);

        let root_actor_ref = kameo::spawn(RootActor::new(
            server_mode,
            fs.clone(),
            net.clone(),
            sleep_duration,
            tx_shutdown,
        ));
        tell!("root", root_actor_ref, root::Start::new(repo)).expect("start root actor");

        info!("Server running - Press Ctrl-C to stop...");
        tokio::select! {
            _r = tokio::signal::ctrl_c() => {
                info!("Ctrl-C received, shutting down...");
            }
            _x = async move {
                    loop{
                        if let Some(message) = rx_shutdown.recv().await {
                            info!("rx shutdown received: {message}");
                            let _ = shutdown_message_holder_clone.write().await.replace(message);
                            tokio::time::sleep(std::time::Duration::from_millis(100)).await;
                            break;
                        }
                    }
                } => {
                    info!("signaled shutdown");
                }
        };
    });

    if let ServerMode::LegacyUI | ServerMode::NewUI = server_mode {
        ratatui::restore();
    }

    if let Some(ref message) = *shutdown_message_holder.blocking_write() {
        info!(%message, "shutdown");
    }

    Ok(())
}

pub fn init_logging() {
    use tracing::Level;
    use tracing_subscriber::FmtSubscriber;

    let subscriber = FmtSubscriber::builder()
        .with_target(false)
        .with_file(true)
        .with_line_number(true)
        .with_max_level(Level::INFO)
        .finish();

    #[allow(clippy::expect_used)]
    tracing::subscriber::set_global_default(subscriber).expect("setting default subscriber failed");
}
