use kameo::message::{Context, Message};
use kx_utils::s;
use ratatui::style::Color;
use tracing::debug;

use crate::{
    server::actor::messages::{RepoUpdate, ServerUpdate},
    tui::{
        actor::{RepoState, ServerState},
        Tui,
    },
};

#[derive(Copy, Clone, Debug)]
pub enum UpdateType {
    Okay,
    Preparation,
    Acting,
    Warning,
}
impl From<UpdateType> for Color {
    fn from(value: UpdateType) -> Self {
        match value {
            UpdateType::Okay => Self::Green,
            UpdateType::Preparation => Self::Gray,
            UpdateType::Acting => Self::LightBlue,
            UpdateType::Warning => Self::Red,
        }
    }
}

impl Message<ServerUpdate> for Tui {
    type Reply = ();

    async fn handle(
        &mut self,
        msg: ServerUpdate,
        _ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        self.state.tap();
        match msg {
            ServerUpdate::AppConfigLoaded { app_config } => {
                self.state.mode = ServerState::from(app_config);
            }

            ServerUpdate::RepoUpdate {
                forge_alias,
                repo_alias,
                repo_update,
            } => {
                if let ServerState::Configured { forges } = &mut self.state.mode {
                    let Some(forge_state) = forges.get_mut(&forge_alias) else {
                        debug!("ServerState::Configured: no forge state available");
                        return;
                    };
                    let Some(repo_state) = forge_state.repos.get_mut(&repo_alias) else {
                        debug!("ServerState::Configured: no repo state available");
                        return;
                    };
                    repo_state.clear_alert();
                    debug!(?repo_update, "ServerState::Configured: RepoUpdate");
                    handle_repo_update(repo_update, repo_state);
                }
            }
        }
    }
}

#[allow(clippy::too_many_lines)]
fn handle_repo_update(repo_update: RepoUpdate, repo_state: &mut RepoState) {
    match repo_update {
        RepoUpdate::Branches { branches } => {
            repo_state.update_branches(branches);
        }
        RepoUpdate::Log { log } => {
            repo_state.update_log(log);
        }
        RepoUpdate::ValidateRepo => {
            repo_state.update_message("polling...", UpdateType::Acting);
        }
        RepoUpdate::Waiting => {
            repo_state.update_message("waiting...", UpdateType::Okay);
            repo_state.clear_alert();
        }
        RepoUpdate::Okay => {
            repo_state.clear_alert();
            repo_state.update_message("okay", UpdateType::Okay);
        }
        RepoUpdate::Alert { alert } => {
            repo_state.alert(alert);
        }
        RepoUpdate::CheckingCI => {
            repo_state.update_message("Checking CI status", UpdateType::Acting);
        }
        RepoUpdate::AdvancingNext { commit, force: _ } => {
            repo_state.update_message(format!("advancing next to {commit}"), UpdateType::Acting);
        }
        RepoUpdate::NextUpdated => {
            repo_state.update_message("next updated - pause while CI starts", UpdateType::Okay);
        }
        RepoUpdate::AdvancingMain { commit } => {
            repo_state.update_message(format!("advancing main to {commit}"), UpdateType::Acting);
        }
        RepoUpdate::AdvanceError(err) => match err {
            crate::repo::AdvanceError::UnconventionalCommitMessage { reason } => repo_state
                .update_message(
                    format!("unconventional commit: {reason}"),
                    UpdateType::Warning,
                ),
            crate::repo::AdvanceError::NextCommitIsWIP => {
                repo_state.update_message(s!("next is a WIP"), UpdateType::Warning);
            }
            crate::repo::AdvanceError::Unexpected(message) => {
                repo_state.update_message(format!("Unexpected: {message}"), UpdateType::Warning);
            }
        },
        RepoUpdate::MainUpdated => {
            repo_state.update_message("main updated", UpdateType::Okay);
        }
        RepoUpdate::Opening { .. } => {
            repo_state.update_message("opening...", UpdateType::Preparation);
        }
        RepoUpdate::Opened => {
            repo_state.update_message("opened", UpdateType::Preparation);
        }
        RepoUpdate::LoadingConfigFromRepo => {
            repo_state.update_message("loading config from repo...", UpdateType::Preparation);
        }
        RepoUpdate::ReceiveCIStatus { status } => {
            repo_state.update_message(format!("ci status: {status:?}"), UpdateType::Warning);
        }
        RepoUpdate::ReceiveRepoConfig { repo_config: _ } => {
            repo_state.update_message("loaded config from repo", UpdateType::Preparation);
        }
        RepoUpdate::RegisteringWebhook => {
            repo_state.update_message("registering webhook...", UpdateType::Preparation);
        }
        RepoUpdate::UnregisteringWebhook => {
            repo_state.update_message("unregistering webhook...", UpdateType::Preparation);
        }
        RepoUpdate::WebhookReceived { push } => {
            repo_state.update_message(
                format!("webhook update: {:?}", push.branch),
                UpdateType::Acting,
            );
        }
        RepoUpdate::RegisteredWebhook => {
            repo_state.update_message("registered webhook", UpdateType::Preparation);
        }
        RepoUpdate::NotOpen => {
            repo_state.update_message("not open", UpdateType::Warning);
        }
        RepoUpdate::NoConfig => {
            repo_state.update_message("no config", UpdateType::Warning);
        }
        RepoUpdate::UserIntervention(user_notification) => {
            repo_state.update_message(user_notification.to_string(), UpdateType::Warning);
        }
        RepoUpdate::Error(message) => {
            repo_state.update_message(format!("error: {message}"), UpdateType::Warning);
        }
        RepoUpdate::ProgressMessage(message) => {
            repo_state.update_message(message, UpdateType::Okay);
        }
        RepoUpdate::RebaseError(error) => {
            repo_state.update_message(error, UpdateType::Warning);
        }
        RepoUpdate::RebaseCompleted => {}
        RepoUpdate::UpdateMessageToken(message_token) => {
            repo_state.update_message_token(message_token);
        }
    }
}
