//
mod handlers;
pub mod messages;
mod model;

#[cfg(test)]
mod tests;

use std::ops::ControlFlow;

use kameo::{
    actor::{ActorRef, WeakActorRef},
    error::{ActorStopReason, PanicError},
    mailbox::unbounded::UnboundedMailbox,
    Actor,
};
use ratatui::{
    crossterm::event::{self, KeyCode, KeyEventKind},
    DefaultTerminal,
};
use tui_scrollview::ScrollViewState;

pub use model::*;

use crate::{tell, IoSendError};

use super::{components::key_focus::KeyFocus, Tick};

#[derive(Debug)]
pub struct Tui {
    terminal: Option<DefaultTerminal>,
    pub state: State,
    scroll_view_state: ScrollViewState,
}
impl Actor for Tui {
    type Mailbox = UnboundedMailbox<Self>;
    type Error = IoSendError;

    async fn on_start(&mut self, actor_ref: ActorRef<Self>) -> Result<(), Self::Error> {
        self.terminal.replace(ratatui::init());
        tell!(actor_ref, Tick)?;
        Ok(())
    }

    async fn on_stop(
        &mut self,
        _actor_ref: WeakActorRef<Self>,
        _reason: ActorStopReason,
    ) -> Result<(), Self::Error> {
        self.terminal.take();
        ratatui::restore();
        Ok(())
    }

    async fn on_panic(
        &mut self,
        _actor_ref: WeakActorRef<Self>,
        err: PanicError,
    ) -> Result<ControlFlow<ActorStopReason>, Self::Error> {
        self.terminal.take();
        ratatui::restore();
        Ok(ControlFlow::Break(ActorStopReason::Panicked(err)))
    }
}
impl Tui {
    pub fn new() -> Self {
        Self {
            terminal: None,
            state: State::initial(),
            scroll_view_state: ScrollViewState::default(),
        }
    }

    fn draw(&mut self) -> std::io::Result<()> {
        if let Some(terminal) = &mut self.terminal {
            terminal.draw(|frame| {
                frame.render_stateful_widget(
                    &self.state,
                    frame.area(),
                    &mut self.scroll_view_state,
                );
            })?;
        } else {
            eprintln!("No terminal setup");
        }
        Ok(())
    }

    fn handle_input(&mut self, actor_tui: &ActorRef<Self>) -> std::io::Result<()> {
        if event::poll(std::time::Duration::from_millis(16))? {
            let event::Event::Key(key) = event::read()? else {
                return Ok(());
            };
            if key.kind != KeyEventKind::Press {
                return Ok(());
            }
            match self.state.key_focus {
                KeyFocus::None => match key.code {
                    KeyCode::Char('q') => {
                        actor_tui.kill();
                    }
                    KeyCode::Char('j') | KeyCode::Down => self.scroll_view_state.scroll_down(),
                    KeyCode::Char('k') | KeyCode::Up => self.scroll_view_state.scroll_up(),
                    KeyCode::Char('f') | KeyCode::PageDown => {
                        self.scroll_view_state.scroll_page_down();
                    }
                    KeyCode::Char('b') | KeyCode::PageUp => {
                        self.scroll_view_state.scroll_page_up();
                    }
                    KeyCode::Char('g') | KeyCode::Home => {
                        self.scroll_view_state.scroll_to_top();
                    }
                    KeyCode::Char('G') | KeyCode::End => {
                        self.scroll_view_state.scroll_to_bottom();
                    }
                    KeyCode::Char('/') => {
                        self.state.key_focus = KeyFocus::Filter;
                    }
                    _ => (),
                },
                KeyFocus::Filter => match key.code {
                    KeyCode::Char(char) => self.state.filter.push(char),
                    KeyCode::Backspace => {
                        self.state.filter.pop();
                    }
                    KeyCode::Tab | KeyCode::Enter => self.state.key_focus = KeyFocus::None,
                    KeyCode::Esc => {
                        self.state.filter = UIRepoFilter::default();
                        self.state.key_focus = KeyFocus::None;
                    }
                    _ => (),
                },
            }
        }
        Ok(())
    }
}
