//
use std::{collections::BTreeMap, fmt::Debug, time::Instant};

use ratatui::{
    layout::Alignment,
    prelude::{Buffer, Rect},
    style::{Color, Style, Stylize as _},
    symbols::border,
    text::{Line, Span},
    widgets::{Block, Paragraph, StatefulWidget, Widget},
};
use tracing::info;
use tui_scrollview::ScrollViewState;

use crate::{
    core::{
        git::{self, graph::Log, Commit},
        ForgeAlias, RepoAlias, RepoBranches,
    },
    repo::messages::MessageToken,
};
use crate::{newtype, s};

use crate::{
    server::actor::messages::ValidAppConfig,
    tui::components::{key_focus::KeyFocus, ConfiguredAppWidget},
};

use super::handlers::UpdateType;

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct State {
    last_update: Instant,
    started: Instant,
    pub mode: ServerState,
    pub filter: UIRepoFilter,
    pub key_focus: KeyFocus,
}
impl State {
    pub fn initial() -> Self {
        Self {
            last_update: Instant::now(),
            started: Instant::now(),
            mode: ServerState::Initial { tick: 0 },
            filter: UIRepoFilter::default(),
            key_focus: KeyFocus::default(),
        }
    }

    pub fn tap(&mut self) {
        self.last_update = Instant::now();
        if let ServerState::Initial { tick } = &mut self.mode {
            *tick += 1;
        }
    }

    fn beating_heart(&self) -> String {
        if self.last_update.duration_since(self.started).as_secs() % 2 == 0 {
            "💚 "
        } else {
            " 💚"
        }
        .to_string()
    }
}

fn time() -> String {
    chrono::Local::now().format("%H:%M").to_string()
}

newtype!(
    UIRepoFilter,
    String,
    Default,
    derive_more::Display,
    "Filter to limit repos shown"
);

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ServerState {
    /// UI has started but has no information on the state of the server
    Initial { tick: usize }, // NOTE: for use with throbber-widgets-tui ?

    /// The application configuration has been loaded, individual forges and repos have their own
    /// states
    Configured {
        forges: BTreeMap<ForgeAlias, ForgeState>,
    },
}
impl ServerState {
    pub fn update_branches(
        &mut self,
        forge_alias: &ForgeAlias,
        repo_alias: &RepoAlias,
        branches: RepoBranches,
    ) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Configured {
                    branches: state_branches,
                    ..
                }
                | RepoState::Ready {
                    branches: state_branches,
                    ..
                } => *state_branches = branches,

                RepoState::Identified { .. } => (),
            }
        }
    }

    pub fn update_log(&mut self, forge_alias: &ForgeAlias, repo_alias: &RepoAlias, log: Log) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Ready { log: state_log, .. } => *state_log = log,

                RepoState::Identified { .. } | RepoState::Configured { .. } => (),
            }
        }
    }
}
impl From<ValidAppConfig> for ServerState {
    fn from(app_config: ValidAppConfig) -> Self {
        Self::Configured {
            forges: app_config
                .app_config
                .forges()
                .map(|(forge_alias, config)| {
                    (
                        forge_alias,
                        config
                            .repos
                            .iter()
                            .map(|(repo_alias, server_repo_config)| {
                                (repo_alias, server_repo_config.repo_config())
                            })
                            .map(|(repo_alias, option_repo_config)| {
                                option_repo_config.map_or_else(
                                    || {
                                        (
                                            repo_alias.clone(),
                                            RepoState::Identified {
                                                repo_alias: repo_alias.clone(),
                                                message: RepoMessage::builder()
                                                    .text("identified".into())
                                                    .style(Style::default().fg(Color::Gray))
                                                    .build(),
                                                alert: None,
                                            },
                                        )
                                    },
                                    |rc| {
                                        (
                                            repo_alias.clone(),
                                            RepoState::Configured {
                                                repo_alias: repo_alias.clone(),
                                                message: RepoMessage::builder()
                                                    .text("configured".into())
                                                    .style(Style::default().fg(Color::LightGreen))
                                                    .build(),
                                                alert: None,
                                                branches: rc.branches().clone(),
                                                log: git::graph::Log::default(),
                                            },
                                        )
                                    },
                                )
                            })
                            .collect::<Vec<_>>(),
                    )
                })
                .map(|(forge_alias, vec_repo_alias_state)| {
                    let forge_state: ForgeState = ForgeState {
                        alias: forge_alias.clone(),
                        view_state: ViewState::default(),
                        repos: vec_repo_alias_state.into_iter().collect::<BTreeMap<_, _>>(),
                    };
                    (forge_alias, forge_state)
                })
                .collect::<BTreeMap<_, _>>(),
        }
    }
}

#[derive(Copy, Clone, Debug, Default, PartialEq, Eq)]
pub enum ViewState {
    Collapsed,
    #[default]
    Expanded,
}
impl std::fmt::Display for ViewState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let view_state = match self {
            Self::Collapsed => "+",
            Self::Expanded => "-",
        };
        write!(f, "{view_state}")
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct ForgeState {
    pub alias: ForgeAlias,
    pub view_state: ViewState,
    pub repos: BTreeMap<RepoAlias, RepoState>,
}

#[derive(bon::Builder, Clone, Debug, PartialEq, Eq)]
pub struct RepoMessage {
    text: String,
    style: Style,
}
impl From<&RepoMessage> for Span<'_> {
    fn from(value: &RepoMessage) -> Self {
        Self::default()
            .content(value.text.clone())
            .style(value.style)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum RepoState {
    Identified {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
    },
    Configured {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        log: Log,
    },
    Ready {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        view_state: ViewState,
        main: Commit,
        next: Commit,
        dev: Commit,
        log: Log,
        message_token: MessageToken,
    },
}
impl RepoState {
    pub const fn repo_alias(&self) -> &RepoAlias {
        match self {
            Self::Identified { repo_alias, .. }
            | Self::Configured { repo_alias, .. }
            | Self::Ready { repo_alias, .. } => repo_alias,
        }
    }

    #[tracing::instrument]
    pub fn update_branches(&mut self, branches: RepoBranches) {
        match self {
            Self::Configured {
                branches: state_branches,
                ..
            }
            | Self::Ready {
                branches: state_branches,
                ..
            } => {
                *state_branches = branches;
            }

            Self::Identified { .. } => (),
        }
    }

    #[tracing::instrument]
    pub fn update_log(&mut self, log: Log) {
        match self {
            Self::Configured { log: state_log, .. } | Self::Ready { log: state_log, .. } => {
                *state_log = log;
            }

            Self::Identified { .. } => {
                info!("git graph log ignored by ui");
            }
        }
    }

    #[tracing::instrument]
    pub fn update_message(
        &mut self,
        msg: impl Into<String> + std::fmt::Debug,
        update_type: UpdateType,
    ) {
        match self {
            Self::Identified { message, .. }
            | Self::Configured { message, .. }
            | Self::Ready { message, .. } => {
                info!(?msg, "updating ui");
                *message = RepoMessage::builder()
                    .text(msg.into())
                    .style(Style::default().fg(update_type.into()))
                    .build();
            }
        }
    }

    #[tracing::instrument]
    pub fn clear_alert(&mut self) {
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => {
                *alert = None;
            }
        }
    }

    #[tracing::instrument]
    pub fn alert(&mut self, msg: impl Into<String> + std::fmt::Debug) {
        let msg: String = msg.into();
        tracing::info!(%msg, "new tui alert");
        self.update_message("ALERT", UpdateType::Warning);
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => *alert = Some(msg),
        }
    }

    pub fn update_message_token(&mut self, new_message_token: MessageToken) {
        if let Self::Ready { message_token, .. } = self {
            *message_token = new_message_token;
        }
    }
}

impl StatefulWidget for &State {
    type State = ScrollViewState;
    fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State)
    where
        Self: Sized,
    {
        let block = Block::bordered()
            .title_top(
                Line::from(
                    if self.filter.is_empty() && self.key_focus == KeyFocus::None {
                        s!("")
                    } else {
                        format!(" Filter: {} ", self.filter)
                    },
                )
                .alignment(Alignment::Left)
                .style(if self.key_focus == KeyFocus::Filter {
                    Color::Red
                } else {
                    Color::Green
                }),
            )
            .title_top(
                Line::from(format!(" Git-Next v{} ", clap::crate_version!()).bold())
                    .alignment(Alignment::Center),
            )
            .title_bottom(
                Line::from(if self.key_focus == KeyFocus::Filter {
                    s!(" [esc] clear [tab/enter] finish ")
                } else {
                    s!("")
                })
                .alignment(Alignment::Left),
            )
            .title_bottom(
                Line::from(vec![
                    " [q]uit [/] filter ".into(),
                    self.beating_heart().into(),
                    " ".into(),
                ])
                .alignment(Alignment::Center),
            )
            .title_bottom(Line::from(format!(" {} ", time())).alignment(Alignment::Right))
            .border_set(border::THICK);
        let interior = block.inner(area);
        block.render(area, buf);
        match &self.mode {
            ServerState::Initial { tick } => Paragraph::new(format!("Loading...{tick}"))
                .centered()
                .render(interior, buf),
            ServerState::Configured { forges } => {
                ConfiguredAppWidget {
                    forges,
                    filter: &self.filter,
                }
                .render(interior, buf, state);
            }
        }
    }
}
