//
mod identity;

use std::string::String;

use crate::core::{RepoAlias, RepoBranches};

use crate::{
    git,
    tui::{
        actor::{RepoMessage, RepoState},
        components::CommitLog,
    },
};

use identity::Identity;

use ratatui::{
    buffer::Buffer,
    layout::Rect,
    widgets::{Block, Borders, Widget},
};

use super::HeightContraintLength;

pub struct RepoWidget<'a> {
    pub repo_state: &'a RepoState,
}
impl HeightContraintLength for RepoWidget<'_> {
    fn height_constraint_length(&self) -> u16 {
        self.inner().height_constraint_length() + 2 // top + bottom borders
    }
}
impl RepoWidget<'_> {
    fn inner(&self) -> InnerRepoWidget {
        match self.repo_state {
            RepoState::Identified {
                repo_alias,
                message,
                alert,
            } => InnerRepoWidget {
                repo_alias,
                message,
                alert: alert.as_ref().map(String::as_str),
                branches: None,
                log: None,
            },

            RepoState::Configured {
                repo_alias,
                message,
                alert,
                branches,
                log,
            }
            | RepoState::Ready {
                repo_alias,
                message,
                alert,
                branches,
                log,
                ..
            } => InnerRepoWidget {
                repo_alias,
                message,
                alert: alert.as_ref().map(String::as_str),
                branches: Some(branches),
                log: Some(log),
            },
        }
    }
}
impl Widget for RepoWidget<'_> {
    fn render(self, area: Rect, buf: &mut Buffer)
    where
        Self: Sized,
    {
        self.inner().render(area, buf);
    }
}

struct InnerRepoWidget<'a> {
    pub repo_alias: &'a RepoAlias,
    pub message: &'a RepoMessage,
    pub alert: Option<&'a str>,
    pub branches: Option<&'a RepoBranches>,
    pub log: Option<&'a git::graph::Log>,
}
impl HeightContraintLength for InnerRepoWidget<'_> {
    fn height_constraint_length(&self) -> u16 {
        self.log
            .map(|log| CommitLog { log })
            .map_or(0, |w| w.height_constraint_length())
    }
}
impl Widget for InnerRepoWidget<'_> {
    fn render(self, area: Rect, buf: &mut Buffer)
    where
        Self: Sized,
    {
        let block = Block::default()
            .title(Identity::new(
                self.repo_alias,
                self.alert,
                self.message,
                self.branches,
            ))
            .borders(Borders::TOP);
        if let Some(log) = self.log {
            CommitLog { log }.render(block.inner(area), buf);
        }
        block.render(area, buf);
    }
}
