//
use std::collections::BTreeMap;

use derive_more::Constructor;
use kameo::{
    actor::ActorRef,
    message::{Context, Message},
    Actor,
};
use tracing::{debug, warn};

use crate::{
    core::{ForgeAlias, RepoAlias},
    repo::{messages::WebhookNotification, RepoActor},
    tell,
};

#[derive(Actor)]
pub struct WebhookRouterActor {
    span: tracing::Span,
    recipients: BTreeMap<ForgeAlias, BTreeMap<RepoAlias, ActorRef<RepoActor>>>,
}
impl Default for WebhookRouterActor {
    fn default() -> Self {
        let span = tracing::info_span!("WebhookRouter");
        Self {
            span,
            recipients: BTreeMap::default(),
        }
    }
}
impl Message<WebhookNotification> for WebhookRouterActor {
    type Reply = color_eyre::Result<()>;

    async fn handle(
        &mut self,
        msg: WebhookNotification,
        _ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        let _gaurd = self.span.enter();
        let forge_alias = msg.forge_alias();
        let repo_alias = msg.repo_alias();
        debug!(forge = %forge_alias, repo = %repo_alias, "Router...");
        let Some(forge_repos) = self.recipients.get(forge_alias) else {
            warn!(forge = %forge_alias, "No forge repos found");
            return Ok(());
        };
        let Some(recipient) = forge_repos.get(repo_alias) else {
            debug!(forge = %forge_alias, repo = %repo_alias, "No recipient found");
            return Ok(());
        };
        tell!(recipient, msg)?;
        Ok(())
    }
}

#[derive(Constructor, Debug)]
pub struct AddWebhookRecipient {
    pub forge_alias: ForgeAlias,
    pub repo_alias: RepoAlias,
    pub recipient: ActorRef<RepoActor>,
}
impl Message<AddWebhookRecipient> for WebhookRouterActor {
    type Reply = ();

    async fn handle(
        &mut self,
        msg: AddWebhookRecipient,
        _ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        if !self.recipients.contains_key(&msg.forge_alias) {
            self.recipients
                .insert(msg.forge_alias.clone(), BTreeMap::new());
        }
        self.recipients
            .get_mut(&msg.forge_alias)
            .map(|repos| repos.insert(msg.repo_alias, msg.recipient));
    }
}
