//
use std::{collections::BTreeMap, net::SocketAddr};

use color_eyre::Result;
use kameo::actor::ActorRef;
use tracing::{info, warn};

use crate::core::{webhook, ForgeAlias, ForgeNotification, RepoAlias};

use crate::{repo::messages::WebhookNotification, tell};

use super::router::WebhookRouterActor;

pub async fn start(socket_addr: SocketAddr, address: ActorRef<WebhookRouterActor>) -> Result<()> {
    // start webhook server
    use warp::Filter;
    // Define the Warp route to handle incoming HTTP requests
    let route = warp::post()
        .map(move || address.clone())
        .and(warp::path::param())
        .and(warp::path::param())
        .and(warp::header::headers_cloned())
        .and(warp::body::bytes())
        .and_then(
            |recipient: ActorRef<WebhookRouterActor>,
             forge_alias: String,
             repo_alias: String,
             // query: String,
             headers: warp::http::HeaderMap,
             body: bytes::Bytes| async move {
                info!("POST received");
                let msg = WebhookNotification::new(ForgeNotification::new(
                    ForgeAlias::new(forge_alias),
                    RepoAlias::new(repo_alias),
                    headers
                        .into_iter()
                        .filter_map(|(k, v)| {
                            k.map(|k| (k.to_string(), v.to_str().unwrap_or_default().to_string()))
                        })
                        .collect::<BTreeMap<String, String>>(),
                    webhook::forge_notification::Body::new(
                        String::from_utf8_lossy(&body).to_string(),
                    ),
                ));
                tell!(recipient, msg)
                    .map(|()| crate::s!("OK"))
                    .map(|reply| {
                        warp::reply::with_header(
                            reply,
                            "Server",
                            format!("git-next {}", clap::crate_version!()),
                        )
                    })
                    .map(|reply| warp::reply::with_status(reply, warp::http::StatusCode::OK))
                    .map_err(|e| {
                        warn!("Unknown error: {:?}", e);
                        warp::reject()
                    })
            },
        );

    // Start the server
    info!("Starting webhook server: {}", socket_addr);
    let (socket, f) = warp::serve(route).try_bind_ephemeral(socket_addr)?;
    info!("Server bound to {socket}");
    f.await;

    Ok(())
}
