use anyhow::Result;
use kameo::{
    actor::{pubsub::PubSub, ActorRef},
    mailbox::unbounded::UnboundedMailbox,
    Actor,
};
use ratatui::DefaultTerminal;
use state::XuiState;
use user_input::{UserInput, UserInputActor};

use crate::{
    base_actor::BaseActor, default_on_actor_link_died, default_on_actor_panic,
    default_on_actor_stop, on_actor_start, subscribe, IoSendError, MessageBus,
};

mod server_update_handler;
mod state;
mod user_input;
mod widgets;

#[derive(Debug, Default)]
pub struct Xui {
    terminal: Option<DefaultTerminal>,
    pub state: XuiState,
    user_input_bus: Option<MessageBus<UserInput>>,
    user_input_actor_ref: Option<ActorRef<UserInputActor>>,
}
impl Actor for Xui {
    type Mailbox = UnboundedMailbox<Self>;
    type Error = IoSendError;

    on_actor_start!(this, actor_ref, {
        this.terminal.replace(ratatui::init());
        let user_input_bus = this.start_user_input_bus(actor_ref.clone()).await;
        this.start_user_input_actor(actor_ref, user_input_bus)
            .await?;
        Ok(())
    });

    default_on_actor_panic!(this, actor_ref, err);
    default_on_actor_link_died!(this, actor_ref, id, reason);
    default_on_actor_stop!(this, actor_ref, reason);
}

impl Xui {
    fn draw(&mut self) -> std::io::Result<()> {
        if let Some(terminal) = &mut self.terminal {
            terminal.draw(|frame| {
                frame.render_stateful_widget(&self.state.data, frame.area(), &mut self.state.ui);
            })?;
        } else {
            eprintln!("No terminal setup");
        }
        tracing::debug!(ui_focus = ?self.state.ui.focus());
        Ok(())
    }

    async fn start_user_input_bus(
        &mut self,
        xui_actor_ref: ActorRef<Self>,
    ) -> MessageBus<UserInput> {
        let actor_ref = PubSub::<UserInput>::new().spawn(xui_actor_ref).await;
        self.user_input_bus.replace(actor_ref.clone());
        actor_ref
    }

    async fn start_user_input_actor(
        &mut self,
        xui_actor_ref: ActorRef<Self>,
        user_input_bus: MessageBus<UserInput>,
    ) -> Result<ActorRef<UserInputActor>> {
        let actor_ref = UserInputActor::builder()
            .user_input_bus(user_input_bus.clone())
            .build()
            .spawn_in_thread(xui_actor_ref.clone())
            .await;
        self.user_input_actor_ref.replace(actor_ref.clone());
        subscribe!(user_input_bus, xui_actor_ref)?;
        Ok(actor_ref)
    }
}
