use derive_with::With;
use kameo::actor::ActorRef;
use ratatui::style::{Style, Stylize};

use crate::{
    core::{
        git::{self, forge::commit::Status, push::Force, Commit},
        webhook::Push,
        RepoBranches,
    },
    repo::{messages::MessageToken, AdvanceError, RepoActor},
};

#[derive(Clone, Debug, Default, With)]
pub struct RepoState {
    repo_actor: Option<ActorRef<RepoActor>>,
    message_token: MessageToken,
    repo_branches: RepoBranches,
    log: git::graph::Log,
    status: RepoStateStatus,
    ci_status: Option<CIStatus>,
}
impl RepoState {
    pub const fn status(&self) -> &RepoStateStatus {
        &self.status
    }

    pub const fn repo_branches(&self) -> &RepoBranches {
        &self.repo_branches
    }

    pub const fn log(&self) -> &git::graph::Log {
        &self.log
    }

    pub const fn ci_status(&self) -> Option<&CIStatus> {
        self.ci_status.as_ref()
    }

    pub const fn repo_actor(&self) -> Option<&ActorRef<RepoActor>> {
        self.repo_actor.as_ref()
    }

    pub const fn message_token(&self) -> MessageToken {
        self.message_token
    }
}

#[derive(
    Clone, Debug, derive_more::Display, Default, PartialEq, Eq, serde::Serialize, serde::Deserialize,
)]
pub enum CIStatus {
    #[default]
    Unknown,
    Okay,
    Checking,
    Received(Status),
}

#[derive(
    Clone, Debug, derive_more::Display, Default, PartialEq, Eq, serde::Serialize, serde::Deserialize,
)]
pub enum RepoStateStatus {
    #[default]
    Unknown,
    Validate,
    #[display("! {alert}")]
    Alert {
        alert: String,
    },
    #[display("Checking CI...")]
    CheckingCI,
    #[display("Advancing Next")]
    AdvancingNext {
        commit: Commit,
        force: Force,
    },
    #[display("Advancing Main")]
    AdvancingMain {
        commit: Commit,
    },
    AdvanceError(AdvanceError),
    Opening,
    #[display("Loading Repo Config")]
    LoadingConfigFromRepo,
    #[display("CI status: {status}")]
    ReceiveCIStatus {
        status: Status,
    },
    #[display("Received Repo Config")]
    ReceiveRepoConfig,
    #[display("Registering Webhook")]
    RegisteringWebhook,
    #[display("Unregistering Webhook")]
    UnregisteringWebhook,
    #[display("Webhook Received: {}", push.branch)]
    WebhookReceived {
        push: Push,
    },
    #[display("Registered Webhook")]
    RegisteredWebhook,
    Opened,
    #[display("Next Updated")]
    NextUpdated,
    #[display("Main Updated")]
    MainUpdated,
    #[display("! {error}")]
    Error {
        error: String,
    },
    Okay,
    UserIntervention {
        notification: crate::core::git::UserNotification,
    },
}
impl RepoStateStatus {
    pub fn style(&self) -> Style {
        let red = Style::default().red();
        let amber = Style::default().yellow();
        let green = Style::default().green();

        match self {
            Self::Unknown | Self::RegisteredWebhook | Self::Opened | Self::Okay => green,
            Self::Alert { .. } | Self::Error { .. } | Self::UserIntervention { .. } => red,
            Self::AdvanceError(err) => match err {
                AdvanceError::UnconventionalCommitMessage { .. } | AdvanceError::Unexpected(_) => {
                    red
                }
                AdvanceError::NextCommitIsWIP => amber,
            },
            Self::Validate
            | Self::Opening
            | Self::CheckingCI
            | Self::AdvancingNext { .. }
            | Self::AdvancingMain { .. }
            | Self::LoadingConfigFromRepo
            | Self::ReceiveRepoConfig
            | Self::RegisteringWebhook
            | Self::UnregisteringWebhook
            | Self::WebhookReceived { .. }
            | Self::MainUpdated
            | Self::NextUpdated => amber,
            Self::ReceiveCIStatus { status } => match status {
                Status::Fail | Status::Error(_) => red,
                Status::Pending => amber,
                Status::Pass => green,
            },
        }
    }
}
#[cfg(test)]
mod test {
    use super::RepoStateStatus;

    #[test]
    fn display() {
        let status = RepoStateStatus::Unknown;

        let _str: String = status.to_string();
    }
}
