use crossterm::event::KeyEvent;
use kameo::message::{Context, Message};

use crate::{
    message,
    xui::{state::UiFocus, Xui},
};

use super::{
    dev_rebase_confirm_handler::handle_dev_rebase_key_event,
    filter_input_handler::handle_filter_input_key_event,
    repo_list_handler::handle_repo_list_key_event,
};

#[derive(Clone, Debug, PartialEq, Eq, serde::Serialize, serde::Deserialize)]
pub enum UserInputEnum {
    KeyEvent(KeyEvent),
    Resize,
}

message!(UserInput, UserInputEnum, "Input from the User");

impl Message<UserInput> for Xui {
    type Reply = color_eyre::Result<()>;

    async fn handle(
        &mut self,
        msg: UserInput,
        ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        match msg.0 {
            UserInputEnum::KeyEvent(key_event) => {
                tracing::debug!(key_code = ?key_event.code);
                match self.state.ui.focus() {
                    UiFocus::RepoList => {
                        handle_repo_list_key_event(key_event, &mut self.state.ui, &ctx.actor_ref())
                            .await?;
                    }
                    UiFocus::FilterInput => {
                        handle_filter_input_key_event(key_event, &mut self.state.ui);
                    }
                    UiFocus::DevRebaseConfirm => {
                        handle_dev_rebase_key_event(key_event, &mut self.state).await?;
                    }
                    UiFocus::RebaseCompleted | UiFocus::RebaseError(_) | UiFocus::Progress(_) => {
                        // any input clears the message
                        self.state.ui.set_focus(UiFocus::RepoList);
                    }
                }
            }
            UserInputEnum::Resize => {} // do nothing, but allow the ui to redraw
        }

        self.draw()?;

        Ok(())
    }
}
