use std::string::ToString;

use ratatui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Style, Stylize as _},
    symbols::border,
    text::{Line, Span},
    widgets::{Block, StatefulWidget, Widget as _},
};

use crate::xui::{
    state::{DataState, RepoKey, RepoState, UiFocus, UiState},
    widgets::Popup,
};

use super::{RepoDetails, RepoTable, RepoTableItem};

impl StatefulWidget for &DataState {
    type State = UiState;

    /// ┏━━━━━━━━━━━━ Git-Next v2025.2.1 ━━━━━━━━━━━━━━━━━┓
    /// ┃ Forge/Repo  Status      ┏━ Log; jo/git-next ━━━┓┃
    /// ┃                         ┃* (dev) foo           ┃┃
    /// ┃ jo/skip     dev-on-main ┃* (next) testing      ┃┃
    /// ┃>jo/git-next CI Running  ┃* (main) fixed        ┃┃
    /// ┃ jo/kxio     CI Failed   ┃                      ┃┃
    /// ┃                         ┃                      ┃┃
    /// ┃                         ┗━━━━━━━━━━━━━━━━━━━━━━┛┃
    /// ┗ \[q]uit [/] filter (jo) ━━━━━━━━━━━━━━━━━━━━━━━━┛
    #[allow(clippy::too_many_lines)]
    fn render(
        self,
        area: ratatui::prelude::Rect,
        buf: &mut ratatui::prelude::Buffer,
        ui_state: &mut Self::State,
    ) {
        // Frame
        let interior = {
            let block = Block::bordered()
                .title_top(
                    Line::from(format!(" Git-Next v{} ", clap::crate_version!()).bold())
                        .alignment(Alignment::Center),
                )
                .title_bottom(key_hints(ui_state).alignment(Alignment::Left))
                .border_set(border::THICK);
            let interior = block.inner(area);
            block.render(area, buf);
            interior
        };

        let (filtered_repos, selected_repo) = self.filtered_and_selected_repos(ui_state);

        let repo_table_items = filtered_repos
            .iter()
            .map(|(repo_key, repo_state)| {
                RepoTableItem::builder()
                    .forge_repo(repo_key.to_string())
                    .status(repo_state.status().to_string())
                    .style(repo_state.status().style())
                    .build()
            })
            .collect::<Vec<_>>();

        let [table_area, repo_area] = Layout::default()
            .direction(Direction::Horizontal)
            .constraints([
                Constraint::Length(repo_table_width(&repo_table_items)),
                Constraint::Fill(1),
            ])
            .areas(interior);

        // List of Repos
        RepoTable::builder().items(repo_table_items).build().render(
            table_area,
            buf,
            ui_state.table_state_mut(),
        );

        // Selected Repo Details
        if let Some(repo) = selected_repo {
            RepoDetails::from(repo).render(repo_area, buf);
        }

        // Popups
        render_popups(area, buf, ui_state);
    }
}

type ForgeRepoItem<'a> = (&'a RepoKey, &'a RepoState);
type ListOfRepos<'a> = Vec<ForgeRepoItem<'a>>;
type ReposSelected<'a> = (ListOfRepos<'a>, Option<ForgeRepoItem<'a>>);

impl DataState {
    fn filtered_and_selected_repos(&self, ui_state: &UiState) -> ReposSelected {
        let filtered_repos = self
            .repos()
            .into_iter()
            .filter(|(repo_key, _)| repo_key.to_string().contains(ui_state.filter()))
            .collect::<Vec<_>>();
        let selected_repo = ui_state
            .table_state()
            .selected()
            .and_then(|selection| filtered_repos.get(selection).copied());
        (filtered_repos, selected_repo)
    }
}

fn repo_table_width(repo_table_items: &[RepoTableItem]) -> u16 {
    #[allow(clippy::cast_possible_truncation)]
    let forge_repo_width: u16 = repo_table_items
        .iter()
        .map(|repo_table_item| repo_table_item.forge_repo().len().clamp(11, 20) as u16)
        .max()
        .unwrap_or(11);
    let status_width = 20;
    let padding = 4;
    // indicator + pre, mid, and post padding
    forge_repo_width + status_width + padding
}

fn render_popups(area: Rect, buf: &mut ratatui::prelude::Buffer, ui_state: &UiState) {
    match &ui_state.focus() {
        UiFocus::DevRebaseConfirm => {
            let popup = Popup::default()
                .with_title(" Rebase dev on main? ")
                .with_content("Are you sure?")
                .with_title_bottom("[Y]es, or [N]o")
                .with_style(Style::default())
                .with_border_style(Style::default())
                .with_title_style(Style::default());
            let popup_area = centered_rect(60, 20, area);
            popup.render(popup_area, buf);
        }
        UiFocus::Progress(message) => {
            let popup = Popup::default()
                .with_title(" Progress ")
                .with_content(message.clone())
                .with_title_bottom(" press any key to continue in background ")
                .with_style(Style::default())
                .with_border_style(Style::default())
                .with_title_style(Style::default());
            let popup_area = centered_rect(60, 20, area);
            popup.render(popup_area, buf);
        }
        UiFocus::RebaseError(error) => {
            let popup = Popup::default()
                .with_title(" Rebase Error ")
                .with_content(error.clone())
                .with_title_bottom(" press any key to continue ")
                .with_style(Style::default())
                .with_border_style(Style::default())
                .with_title_style(Style::default().red());
            let popup_area = centered_rect(60, 20, area);
            popup.render(popup_area, buf);
        }
        UiFocus::RebaseCompleted => {
            let popup = Popup::default()
                .with_title(" Rebase Complete ")
                .with_content("dev branch has been rebased onto main and pushed")
                .with_title_bottom(" press any key to continue ")
                .with_style(Style::default())
                .with_border_style(Style::default())
                .with_title_style(Style::default().red());
            let popup_area = centered_rect(60, 20, area);
            popup.render(popup_area, buf);
        }
        UiFocus::RepoList | UiFocus::FilterInput => (),
    }
}

/// helper function to create a centered rect using up certain percentage of the available rect `r`
fn centered_rect(percent_x: u16, percent_y: u16, r: Rect) -> Rect {
    let popup_layout = Layout::default()
        .direction(Direction::Vertical)
        .constraints([
            Constraint::Percentage((100 - percent_y) / 2),
            Constraint::Percentage(percent_y),
            Constraint::Percentage((100 - percent_y) / 2),
        ])
        .split(r);

    Layout::default()
        .direction(Direction::Horizontal)
        .constraints([
            Constraint::Percentage((100 - percent_x) / 2),
            Constraint::Percentage(percent_x),
            Constraint::Percentage((100 - percent_x) / 2),
        ])
        .split(popup_layout[1])[1]
}

fn key_hints(ui_state: &UiState) -> Line<'_> {
    let mut hints: Vec<Span<'_>> = vec![Span::from(" ")];
    let filter_hint = format!("({})", ui_state.filter());
    match &ui_state.focus() {
        UiFocus::RepoList => {
            hints.extend(vec!["[".into(), "q".red(), "]uit ".into()]);
            hints.extend(vec!["[".into(), "/".red(), "] filter ".into()]);
            if !ui_state.filter().is_empty() {
                hints.push(filter_hint.into());
            }
        }
        UiFocus::FilterInput => {
            hints.push("filter: ".into());
            hints.push(ui_state.filter().into());
            hints.extend(vec![" [".into(), "esc".red(), "] clear ".into()]);
            hints.extend(vec![
                "[".into(),
                "tab".red(),
                "/".into(),
                "enter".red(),
                "] finish ".into(),
            ]);
        }
        UiFocus::DevRebaseConfirm => {
            hints.extend(vec!["[".into(), "y".red(), "]es".into()]);
            hints.extend(vec!["[".into(), "n".red(), "]o".into()]);
        }
        UiFocus::RebaseCompleted | UiFocus::RebaseError(_) | UiFocus::Progress(_) => {}
    };
    Line::from(hints)
}

#[cfg(test)]
mod tests {
    use assert2::let_assert;
    use kx_utils::s;

    use crate::xui::state::UiState;

    use super::*;

    #[test]
    fn no_repo_selected_repo_details_is_none() {
        //given
        let mut data_state = DataState::default();
        let repo_key = RepoKey::builder()
            .forge_alias(s!("jo").into())
            .repo_alias(s!("git-next").into())
            .build();
        let repo_state = RepoState::default();
        data_state.set_repo_state(repo_key, repo_state);

        let mut ui_state = UiState::default();
        ui_state.filter_mut().push_str("abcd"); // matches nothing

        //when
        let (repos, selected) = data_state.filtered_and_selected_repos(&ui_state);

        //then
        assert!(repos.is_empty());
        assert!(selected.is_none());
    }

    #[test]
    fn repo_selected_repo_details_is_some() {
        //given
        let mut data_state = DataState::default();
        let repo_key = RepoKey::builder()
            .forge_alias(s!("jo").into())
            .repo_alias(s!("git-next").into())
            .build();
        let repo_state = RepoState::default();
        data_state.set_repo_state(repo_key.clone(), repo_state.clone());
        data_state.set_repo_state(
            RepoKey::builder()
                .forge_alias(s!("gh").into())
                .repo_alias(s!("skip").into())
                .build(),
            repo_state,
        );

        let mut ui_state = UiState::default();
        ui_state.filter_mut().push_str("jo");
        ui_state.table_state_mut().select_first();

        //when
        let (repos, selected) = data_state.filtered_and_selected_repos(&ui_state);

        //then
        assert_eq!(repos.len(), 1, "filtered from 2 to 1");
        let_assert!(Some((key, _)) = repos.first());
        assert_eq!(**key, repo_key, "list doesn't have the expected repo");
        let_assert!(Some((selection, _)) = selected);
        assert_eq!(*selection, repo_key, "selection isn't the expected repo");
    }
}
