use bon::Builder;
use ratatui::{
    layout::{Alignment, Constraint, Layout},
    prelude::{Buffer, Rect},
    text::Line,
    widgets::{Block, Borders, Widget},
};

use crate::{
    core::{
        git::{graph::Log, UserNotification},
        ForgeAlias, RepoAlias, RepoBranches,
    },
    xui::state::{CIStatus, RepoKey, RepoState, RepoStateStatus},
};

use super::{CommitLog, UserAlert};

#[derive(Builder, Clone, Debug, PartialEq, Eq, serde::Serialize, serde::Deserialize)]
pub struct RepoDetails {
    forge_alias: ForgeAlias,
    repo_alias: RepoAlias,
    repo_branches: RepoBranches,
    log: Log,
    ci_state: Option<CIStatus>,
    status: RepoStateStatus,
}

impl From<(&RepoKey, &RepoState)> for RepoDetails {
    fn from((repo_key, repo_state): (&RepoKey, &RepoState)) -> Self {
        Self {
            forge_alias: repo_key.forge_alias().clone(),
            repo_alias: repo_key.repo_alias().clone(),
            repo_branches: repo_state.repo_branches().clone(),
            log: repo_state.log().clone(),
            ci_state: repo_state.ci_status().cloned(),
            status: repo_state.status().clone(),
        }
    }
}

impl Widget for RepoDetails {
    fn render(self, area: Rect, buf: &mut Buffer)
    where
        Self: Sized,
    {
        let block = Block::default()
            .title(format!(
                " {}/{} {} [{}] ",
                self.forge_alias,
                self.repo_alias,
                self.ci_state.unwrap_or_default(),
                self.status
            ))
            .title_bottom(key_hints(&self.status).alignment(Alignment::Left))
            .borders(Borders::ALL);
        let interior = block.inner(area);
        block.render(area, buf);
        let [notification_area, log_area, branches_area] = Layout::default()
            .direction(ratatui::layout::Direction::Vertical)
            .constraints([
                Constraint::Length(1), // user notification
                Constraint::Fill(1),   // Git Log
                Constraint::Length(1), // repo_branches
            ])
            .areas(interior);

        if let RepoStateStatus::UserIntervention { notification } = &self.status {
            UserAlert::builder()
                .notification(notification)
                .build()
                .render(notification_area, buf);
        }

        CommitLog::builder()
            .log(&self.log)
            .build()
            .render(log_area, buf);

        Line::from(format!(
            " ({} -> {} -> {}) ",
            self.repo_branches.dev(),
            self.repo_branches.next(),
            self.repo_branches.main()
        ))
        .alignment(Alignment::Right)
        .render(branches_area, buf);
    }
}

fn key_hints(status: &RepoStateStatus) -> Line<'_> {
    if let RepoStateStatus::UserIntervention { notification } = status {
        Line::from(match notification {
            UserNotification::CICheckFailed { .. }
            | UserNotification::RepoConfigLoadFailure { .. }
            | UserNotification::WebhookRegistration { .. } => "[r]etry",
            UserNotification::DevNotBasedOnMain { .. } => "[r]ebase",
        })
    } else {
        Line::from("")
    }
}
