//
use std::fmt::Display;

use crate::fs::{DirItem, DirItemIterator, Result};

use super::{DirHandle, Error, PathHandle, PathMarker};

impl DirHandle {
    /// Creates a new, empty directory at the path
    ///
    /// Wrapper for [std::fs::create_dir]
    ///
    /// ```
    /// # fn try_main() -> kxio::fs::Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo");
    /// let dir = fs.dir(&path);
    /// dir.create()?;
    /// # Ok(())
    /// # }
    /// ```
    #[tracing::instrument]
    pub fn create(&self) -> Result<()> {
        self.check_error()?;
        tracing::debug!("std::fs::create_dir");
        std::fs::create_dir(self.as_pathbuf()).map_err(Error::Io)
    }

    /// Recursively create a directory and all of its parent components if they are missing.
    ///
    /// Wrapper for [std::fs::create_dir_all]
    ///
    /// ```
    /// # fn try_main() -> kxio::fs::Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo").join("bar");
    /// let dir = fs.dir(&path);
    /// dir.create_all()?;
    /// # Ok(())
    /// # }
    /// ```
    #[tracing::instrument]
    pub fn create_all(&self) -> Result<()> {
        self.check_error()?;
        tracing::debug!("std::fs::create_dir_all");
        std::fs::create_dir_all(self.as_pathbuf()).map_err(Error::Io)
    }

    /// Returns an iterator over the entries within a directory.
    ///
    /// Wrapper for [std::fs::read_dir]
    ///
    /// ```
    /// # fn try_main() -> kxio::fs::Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo").join("bar");
    /// let dir = fs.dir(&path);
    /// for entry in dir.read()? { /* ... */ }
    /// # Ok(())
    /// # }
    /// ```
    #[tracing::instrument]
    pub fn read(&self) -> Result<Box<dyn Iterator<Item = Result<DirItem>>>> {
        self.check_error()?;
        tracing::debug!("std::fs::read_dir");
        let read_dir = std::fs::read_dir(self.as_pathbuf()).map_err(Error::Io)?;
        Ok(Box::new(DirItemIterator::new(read_dir)))
    }

    /// Removes an empty directory.
    ///
    /// Wrapper for [std::fs::remove_dir]
    ///
    /// ```
    /// # fn try_main() -> kxio::fs::Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo").join("bar");
    /// let dir = fs.dir(&path);
    /// dir.remove()?;
    /// # Ok(())
    /// }
    #[tracing::instrument]
    pub fn remove(&self) -> Result<()> {
        self.check_error()?;
        tracing::debug!("std::fs::remove");
        std::fs::remove_dir(self.as_pathbuf()).map_err(Error::Io)
    }

    /// Recursively remove a directory and all of its contents.
    ///
    /// Wrapper for [std::fs::remove_dir_all]
    ///
    /// ```
    /// # fn try_main() -> kxio::fs::Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo").join("bar");
    /// let dir = fs.dir(&path);
    /// dir.remove_all()?;
    /// # Ok(())
    /// }
    #[tracing::instrument]
    pub fn remove_all(&self) -> Result<()> {
        self.check_error()?;
        tracing::debug!("std::fs::remove_all");
        std::fs::remove_dir_all(self.as_pathbuf()).map_err(Error::Io)
    }
}
impl TryFrom<PathHandle<PathMarker>> for DirHandle {
    type Error = crate::fs::Error;

    fn try_from(path: PathHandle<PathMarker>) -> std::result::Result<Self, Self::Error> {
        match path.as_dir() {
            Ok(Some(dir)) => Ok(dir.clone()),
            Ok(None) => Err(crate::fs::Error::NotADirectory {
                path: path.as_pathbuf(),
            }),
            Err(err) => Err(err),
        }
    }
}

impl Display for DirHandle {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}/", self.as_pathbuf().display())
    }
}
