//
use std::{fmt::Display, path::Path, str::Lines};

use crate::fs::Result;

use super::Error;

/// A reader for a file.
#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct Reader {
    contents: Vec<u8>,
}
impl Reader {
    #[tracing::instrument]
    pub(super) fn new(path: &Path) -> Result<Self> {
        tracing::debug!("std::fs::read_to_string");
        let contents = std::fs::read(path).map_err(Error::Io)?;
        tracing::debug!(len = contents.len(), "contents");
        Ok(Self { contents })
    }

    /// Returns the contents of the file as a string.
    ///
    /// ```
    /// # use kxio::fs::Result;
    /// # fn main() -> Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo");
    /// let file = fs.file(&path);
    /// # file.write("new file contents")?;
    /// let contents = file.reader()?.as_str();
    /// #    Ok(())
    /// # }
    /// ```
    pub fn as_str(&self) -> Result<&str> {
        std::str::from_utf8(&self.contents).map_err(Error::Utf8Error)
    }

    /// Returns an iterator over the lines in the file.
    ///
    /// ```
    /// # use kxio::fs::Result;
    /// # fn main() -> Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo");
    /// let file = fs.file(&path);
    /// # file.write("new file contents")?;
    /// let lines = file.reader()?.lines();
    /// #    Ok(())
    /// # }
    /// ```
    pub fn lines(&self) -> Result<Lines<'_>> {
        self.as_str().map(|s| s.lines())
    }

    /// Returns the contents of the file as bytes.
    ///
    /// ```
    /// # use kxio::fs::Result;
    /// # fn main() -> Result<()> {
    /// let fs = kxio::fs::temp()?;
    /// let path = fs.base().join("foo");
    /// let file = fs.file(&path);
    /// # file.write("new file contents")?;
    /// let bytes = file.reader()?.bytes();
    /// #    Ok(())
    /// # }
    /// ```
    pub fn bytes(&self) -> &[u8] {
        &self.contents
    }
}
impl Display for Reader {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_str().map_err(|_| std::fmt::Error)?)
    }
}
